"""
Methods and / or classes to perform 3D colocation
"""

from __future__ import annotations

import logging
import os
from typing import NamedTuple

import iris
import numpy as np
from cf_units import Unit

from pyaerocom import __version__ as pya_ver
from pyaerocom import const
from pyaerocom._lowlevel_helpers import LayerLimits, RegridResDeg
from pyaerocom.exceptions import (
    DataUnitError,
    DimensionOrderError,
    MetaDataError,
    TemporalResolutionError,
    VarNotAvailableError,
)
from pyaerocom.filter import Filter
from pyaerocom.helpers import make_datetime_index
from pyaerocom.tstype import TsType

from .colocated_data import ColocatedData
from .colocation_utils import (
    _colocate_site_data_helper,
    _colocate_site_data_helper_timecol,
    _regrid_gridded,
    check_time_ival,
    check_ts_type,
    resolve_var_name,
)

logger = logging.getLogger(__name__)


class ColocatedDataLists(NamedTuple):
    colocateddata_for_statistics: list[ColocatedData]
    colocateddata_for_profile_viz: list[ColocatedData]


def _colocate_vertical_profile_gridded(
    data,
    data_ref,
    start=None,
    stop=None,
    filter_name=None,
    harmonise_units=True,
    var_ref=None,
    min_num_obs=None,
    colocate_time=False,
    use_climatology_ref=False,
    resample_how=None,
    layer_limits: LayerLimits | None = None,
    obs_stat_data=None,
    ungridded_lons=None,
    ungridded_lats=None,
    col_freq=None,
    col_tst=None,
    var=None,
    var_aerocom=None,
    var_ref_aerocom=None,
) -> list[ColocatedData]:
    if layer_limits is None:
        raise Exception("layer limits must be provided")

    data_ref_unit = None
    ts_type_src_ref = None
    if not harmonise_units:
        data_unit = str(data.units)
    else:
        data_unit = None

    pd_freq = col_tst.to_pandas_freq()
    time_idx = make_datetime_index(start, stop, pd_freq)

    time_num = len(time_idx)
    stat_num = len(obs_stat_data)

    arr = np.full((2, time_num, stat_num), np.nan)

    lons = np.full(stat_num, np.nan)
    lats = np.full(stat_num, np.nan)
    alts = np.full(stat_num, np.nan)

    station_names = [""] * stat_num

    dataset_ref = data_ref.contains_datasets[0]
    ts_type_src_data = data.ts_type

    list_of_colocateddata_objects = []
    for vertical_layer in layer_limits:
        # Think about efficency here in terms of order of loops. candidate for parallelism
        # create the 2D layer data
        arr = np.full((2, time_num, stat_num), np.nan)
        try:
            data_this_layer = (
                data.extract(
                    iris.Constraint(
                        coord_values={
                            "altitude": lambda cell: vertical_layer["start"]
                            < cell
                            < vertical_layer["end"]
                        }
                    )
                )
                .collapsed("altitude", iris.analysis.MEAN)
                .copy()
            )
        except Exception as e:
            logger.warning(f"No altitude in model data layer {vertical_layer}")
            logger.debug(f"Raised: {e}")
            continue

        grid_stat_data_this_layer = data_this_layer.to_time_series(
            longitude=ungridded_lons,
            latitude=ungridded_lats,
        )

        # loop over all stations and append to colocated data object
        for i, obs_stat in enumerate(obs_stat_data):
            # Add coordinates to arrays required for xarray.DataArray below
            lons[i] = obs_stat.longitude
            lats[i] = obs_stat.latitude
            alts[i] = obs_stat.station_coords[
                "altitude"
            ]  # altitude refers to altitude of the data. be explcit where getting from
            station_names[i] = obs_stat.station_name

            # ToDo: consider removing to keep ts_type_src_ref (this was probably
            # introduced for EBAS were the original data frequency is not constant
            # but can vary from site to site)
            if ts_type_src_ref is None:
                ts_type_src_ref = obs_stat["ts_type_src"]
            elif obs_stat["ts_type_src"] != ts_type_src_ref:
                spl = ts_type_src_ref.split(";")
                if obs_stat["ts_type_src"] not in spl:
                    spl.append(obs_stat["ts_type_src"])
                ts_type_src_ref = ";".join(spl)

            if data_ref_unit is None:
                try:
                    data_ref_unit = obs_stat["var_info"][var_ref]["units"]
                except KeyError as e:  # variable information or unit is not defined
                    logger.exception(repr(e))
            try:
                unit = obs_stat["var_info"][var_ref]["units"]
            except Exception:
                unit = None
            if not unit == data_ref_unit:
                raise ValueError(
                    f"Cannot perform colocation. "
                    f"Ungridded data object contains different units ({var_ref})"
                )
            # get observations (Note: the index of the observation time series
            # is already in the specified frequency format, and thus, does not
            # need to be updated, for details (or if errors occur), cf.
            # UngriddedData.to_station_data, where the conversion happens)

            # get model station data
            grid_stat_this_layer = grid_stat_data_this_layer[i]

            # Make a copy of the station data and resample it to the mean based on hourly resolution. Needs testing!
            obs_stat_this_layer = obs_stat.copy()

            try:
                obs_stat_this_layer[var_ref] = (
                    obs_stat_this_layer.select_altitude(
                        var_name=var_ref, altitudes=list(vertical_layer.values())
                    )
                    .mean("altitude", skipna=True)  # very important to skip nans here
                    .to_series()  # make pandas series
                )
            except ValueError:
                logger.warning(
                    f"Var: {var_ref}. Skipping {obs_stat_this_layer.station_name} in altitude layer {vertical_layer} because no data"
                )
                continue

            if harmonise_units:
                grid_unit = grid_stat_this_layer.get_unit(var)
                obs_unit = obs_stat_this_layer.get_unit(var_ref)
                if not grid_unit == obs_unit:
                    grid_stat_this_layer.convert_unit(var, obs_unit)
                if data_unit is None:
                    data_unit = obs_unit

            try:
                if colocate_time:
                    _df = _colocate_site_data_helper_timecol(
                        stat_data=grid_stat_this_layer,
                        stat_data_ref=obs_stat_this_layer,
                        var=var,
                        var_ref=var_ref,
                        ts_type=col_freq,
                        resample_how=resample_how,
                        min_num_obs=min_num_obs,
                        use_climatology_ref=use_climatology_ref,
                    )
                else:
                    _df = _colocate_site_data_helper(
                        stat_data=grid_stat_this_layer,
                        stat_data_ref=obs_stat_this_layer,
                        var=var,
                        var_ref=var_ref,
                        ts_type=col_freq,
                        resample_how=resample_how,
                        min_num_obs=min_num_obs,
                        use_climatology_ref=use_climatology_ref,
                    )

                try:
                    # assign the unified timeseries data to the colocated data array
                    arr[0, :, i] = _df["ref"].values
                    arr[1, :, i] = _df["data"].values
                except ValueError:
                    try:
                        mask = _df.index.intersection(time_idx)
                        _df = _df.loc[mask]
                        arr[0, :, i] = _df["ref"].values
                        arr[1, :, i] = _df["data"].values
                    except ValueError as e:
                        logger.warning(
                            f"Failed to colocate time for station {obs_stat.station_name}. "
                            f"This station will be skipped (error: {e})"
                        )
            except TemporalResolutionError as e:
                # resolution of obsdata is too low
                logger.warning(
                    f"{var_ref} data from site {obs_stat.station_name} will "
                    f"not be added to ColocatedData. Reason: {e}"
                )

        try:
            revision = data_ref.data_revision[dataset_ref]
        except Exception:
            try:
                revision = data_ref._get_data_revision_helper(dataset_ref)
            except MetaDataError:
                revision = "MULTIPLE"
            except Exception:
                revision = "n/a"

        files = [os.path.basename(x) for x in data.from_files]

        meta = {
            "data_source": [dataset_ref, data.data_id],
            "var_name": [var_ref_aerocom, var_aerocom],
            "var_name_input": [var_ref, var],
            "ts_type": col_freq,  # will be updated below if resampling
            "filter_name": filter_name,
            "ts_type_src": [ts_type_src_ref, ts_type_src_data],
            "var_units": [data_ref_unit, data_unit],
            "data_level": 3,
            "revision_ref": revision,
            "from_files": files,
            "from_files_ref": None,
            "colocate_time": colocate_time,
            "obs_is_clim": use_climatology_ref,
            "pyaerocom": pya_ver,
            "min_num_obs": min_num_obs,
            "resample_how": resample_how,
            "vertical_layer": vertical_layer,
        }

        # create coordinates of DataArray
        coords = {
            "data_source": meta["data_source"],
            "time": time_idx,
            "station_name": station_names,
            "latitude": ("station_name", lats),
            "longitude": ("station_name", lons),
            "altitude": ("station_name", alts),
        }

        dims = ["data_source", "time", "station_name"]
        coldata = ColocatedData(data=arr, coords=coords, dims=dims, name=var, attrs=meta)

        # add correct units for lat / lon dimensions
        coldata.latitude.attrs["standard_name"] = data.latitude.standard_name
        coldata.latitude.attrs["units"] = str(data.latitude.units)

        coldata.longitude.attrs["standard_name"] = data.longitude.standard_name
        coldata.longitude.attrs["units"] = str(data.longitude.units)

        coldata.data.attrs["altitude_units"] = str(data.altitude.units)

        coldata.vertical_layer = vertical_layer

        list_of_colocateddata_objects.append(coldata)

    return list_of_colocateddata_objects


def colocate_vertical_profile_gridded(
    data,
    data_ref,
    ts_type: str = None,
    start: str | None = None,
    stop: str | None = None,
    filter_name: str = None,
    regrid_res_deg: float | RegridResDeg | None = None,
    harmonise_units: bool = True,
    regrid_scheme: str = "areaweighted",
    var_ref: str = None,
    update_baseyear_gridded: int = None,
    min_num_obs: int | dict | None = None,
    colocate_time: bool = False,
    use_climatology_ref: bool = False,
    resample_how: str | dict = None,
    colocation_layer_limits: tuple[LayerLimits, ...] | None = None,
    profile_layer_limits: tuple[LayerLimits, ...] | None = None,
    **kwargs,
) -> ColocatedDataLists:
    """
    Colocated vertical profile data with gridded (model) data

    The guts of this function are placed in a helper function as not to repeat the code.
    This is done because colocation must occur twice:
        i) at the the statistics are computed
        ii) at a finder vertical resoltuion for profile vizualization
    Some things you do not want to compute twice, however.
    So (most of) the things that correspond to both colocation instances are computed here,
    and then passed to the helper function.

    Returns
        colocated_data_lists : ColocatedDataLists

    -------
    """

    if filter_name is None:
        filter_name = const.DEFAULT_REG_FILTER
    try:
        data.check_dimcoords_tseries()
    except DimensionOrderError:
        data.reorder_dimensions_tseries()

    var, var_aerocom = resolve_var_name(data)
    if var_ref is None:
        var_ref = var_aerocom
        var_ref_aerocom = var_aerocom
    else:
        var_ref_aerocom = const.VARS[var_ref].var_name_aerocom

    if var_ref not in data_ref.contains_vars:
        raise VarNotAvailableError(
            f"Variable {var_ref} is not available in ungridded "
            f"data (which contains {data_ref.contains_vars})"
        )
    elif len(data_ref.contains_datasets) > 1:
        raise AttributeError(
            f"Colocation can only be performed with ungridded data objects "
            f"that only contain a single dataset (input data contains: "
            f"{data_ref.contains_datasets}. Use method `extract_dataset` of "
            f"UngriddedData object to extract single datasets."
        )

    if any(
        not {"start", "end"}.issubset(layer)
        for layer in colocation_layer_limits + profile_layer_limits
    ):
        raise KeyError(
            "start and end must be provided for profiles in each vertical layer in colocate_vertical_profile_gridded"
        )

    data_ref_meta_idxs_with_var_info = []
    for i in range(len(data_ref.metadata)):
        if "altitude" not in data_ref.metadata[i]["var_info"]:
            logger.warning(
                f"Warning: Station {data_ref.metadata[i]['station_name']} does not have any var_info"
            )
        else:
            data_ref_meta_idxs_with_var_info.append(i)

    if any(
        data.altitude.units != Unit(data_ref.metadata[i]["var_info"]["altitude"]["units"])
        for i in data_ref_meta_idxs_with_var_info
    ):
        logger.info(
            f"Mismatching units in colocation_3d.py. Model has units {data.altitude.units} whereas not all observations have this unit. Debug to find out where."
        )
        raise DataUnitError

    if update_baseyear_gridded is not None:
        # update time dimension in gridded data
        data.base_year = update_baseyear_gridded

    # apply region filter to data
    regfilter = Filter(name=filter_name)
    data_ref = regfilter.apply(data_ref)
    data = regfilter.apply(data)

    # check time overlap and crop model data if needed
    start, stop = check_time_ival(data, start, stop)
    data = data.crop(time_range=(start, stop))

    if regrid_res_deg is not None:  # pragma: no cover
        data = _regrid_gridded(data, regrid_scheme, regrid_res_deg)

    # Special ts_typs for which all stations with ts_type< are removed
    reduce_station_data_ts_type = ts_type

    # ts_type_src_data = data.ts_type
    ts_type, ts_type_data = check_ts_type(data, ts_type)
    if not colocate_time and ts_type < ts_type_data:
        data = data.resample_time(str(ts_type), min_num_obs=min_num_obs, how=resample_how)
        ts_type_data = ts_type

    if use_climatology_ref:  # pragma: no cover
        col_freq = "monthly"
        obs_start = const.CLIM_START
        obs_stop = const.CLIM_STOP
    else:
        col_freq = str(ts_type)
        obs_start = start
        obs_stop = stop

    # colocation frequency
    col_tst = TsType(col_freq)

    latitude = data.latitude.points
    longitude = data.longitude.points
    altitude = data.altitude.points
    lat_range = [np.min(latitude), np.max(latitude)]
    lon_range = [np.min(longitude), np.max(longitude)]
    alt_range = [np.min(altitude), np.max(altitude)]
    # use only sites that are within model domain

    # filter_by_meta wipes is_vertical_profile
    # Also note that filter_by_meta may not be calling alt_range. Function fitler_altitude is defined but not used
    data_ref = data_ref.filter_by_meta(latitude=lat_range, longitude=lon_range, altitude=alt_range)

    # get timeseries from all stations in provided time resolution
    # (time resampling is done below in main loop)
    all_stats = data_ref.to_station_data_all(
        vars_to_convert=var_ref,
        start=obs_start,
        stop=obs_stop,
        by_station_name=True,
        ts_type_preferred=reduce_station_data_ts_type,
        **kwargs,
    )

    if len(all_stats["stats"]) == 0:
        raise VarNotAvailableError(
            f"Variable {var_ref} is not available in specified time interval ({start}-{stop})"
        )

    # Colocation has to occur twice for vertical profiles.
    # Once for the colocation which we will compute the statistics over.
    # The second time is just to show the vertical profiles on the web. This needs to be finer
    # Here we make a list with the list of ColocatedData objects for both colocation purposes
    output_prep = [
        _colocate_vertical_profile_gridded(
            data=data,
            data_ref=data_ref,
            start=start,
            stop=stop,
            filter_name=filter_name,
            harmonise_units=harmonise_units,
            var_ref=var_ref,
            min_num_obs=min_num_obs,
            colocate_time=colocate_time,
            use_climatology_ref=use_climatology_ref,
            resample_how=resample_how,
            layer_limits=layer_limits,
            obs_stat_data=all_stats["stats"],
            ungridded_lons=all_stats["longitude"],
            ungridded_lats=all_stats["latitude"],
            col_freq=col_freq,
            col_tst=col_tst,
            var=var,
            var_aerocom=var_aerocom,
            var_ref_aerocom=var_ref_aerocom,
        )
        for layer_limits in (colocation_layer_limits, profile_layer_limits)
    ]
    # Create a namedtuple for output.
    # Each element in the tuple is a list of ColocatedData objects.
    # The length of these lists is the same as the number of colocation layers

    for coldata in output_prep[1]:
        coldata.data.attrs["just_for_viz"] = 1

    colocated_data_lists = ColocatedDataLists(
        *output_prep
    )  # put the list of prepared output into namedtuple object s.t. both position and named arguments can be used

    return colocated_data_lists



from __future__ import annotations

from datetime import date, timedelta
from itertools import product
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from pyaerocom import const

from . import cfg_test, cfg_test_mos


@pytest.fixture()
def fake_cache_path(monkeypatch, tmp_path: Path):
    monkeypatch.setattr(
        "pyaerocom.io.cachehandler_ungridded.CacheHandlerUngridded.cache_dir", tmp_path
    )
    cache_file = tmp_path / "tmp.pkl"
    cache_file.write_bytes(b"")
    assert cache_file.exists()
    return tmp_path


@pytest.fixture
def reset_cachedir():
    cache = const.CACHEDIR
    yield
    const.CACHEDIR = cache


@pytest.fixture
def patched_config():
    cfg = cfg_test.CFG
    assert cfg["proj_id"] == "cams2-83"
    return cfg


@pytest.fixture
def patched_config_mos():
    cfg = cfg_test_mos.CFG
    assert cfg["exp_id"] == "mos-colocated-data"
    return cfg


@pytest.fixture
def fake_CAMS2_83_Processer(monkeypatch):
    def do_not_run(
        self,
        model_name=None,
        obs_name=None,
        var_list=None,
        update_interface=True,
        analysis=False,
    ):
        assert model_name is None
        assert obs_name is None
        assert var_list is None
        assert analysis is False
        assert update_interface is True

    monkeypatch.setattr("pyaerocom.scripts.cams2_83.evaluation.CAMS2_83_Processer.run", do_not_run)


@pytest.fixture
def fake_ExperimentProcessor(monkeypatch):
    def do_not_run(self, model_name=None, obs_name=None, var_list=None, update_interface=True):
        assert model_name is None
        assert obs_name is None
        assert var_list is None
        assert update_interface is True

    monkeypatch.setattr(
        "pyaerocom.scripts.cams2_83.evaluation.ExperimentProcessor.run", do_not_run
    )


@pytest.fixture(scope="module")
def coldata_mos(tmp_path_factory) -> Path:
    root: Path = tmp_path_factory.mktemp("data")

    def dataset(model: str, day: int, start: date, end: date) -> xr.Dataset:
        hours = (end - start) // timedelta(hours=1) + 1
        ds = xr.Dataset(
            data_vars=dict(
                concno2=xr.Variable(
                    ("data_source", "time", "station_name"),
                    np.zeros((2, hours, 3)),
                    {
                        "ts_type": "hourly",
                        "filter_name": "ALL-wMOUNTAINS",
                        "ts_type_src": ["hourly", "hourly"],
                        "var_units": ["ug m-3", "ug m-3"],
                        "data_level": 3,
                        "revision_ref": "n/a",
                        "from_files": "",
                        "from_files_ref": "None",
                        "colocate_time": 0,
                        "obs_is_clim": 0,
                        "pyaerocom": "0.18.dev0",
                        "CONV!min_num_obs": str(dict(daily=dict(hourly=18))),
                        "resample_how": "None",
                        "obs_name": "EEA-UTD",
                        "vert_code": "Surface",
                        "diurnal_only": 0,
                        "zeros_to_nan": 1,
                    },
                )
            ),
            coords=dict(
                data_source=xr.Variable(
                    "data_source", ["CAMS2_83.NRT", f"CAMS2-83.{model}.day{day}.FC"]
                ),
                station_name=xr.Variable("station_name", ["AT0ENK1", "AT0ILL1", "XK0012A"]),
                latitude=xr.Variable("station_name", [48.39, 47.77, 42.66]),
                longitude=xr.Variable("station_name", [13.67, 16.77, 21.08]),
                altitude=xr.Variable("station_name", [525, 117, 529]),
                time=xr.Variable("time", pd.date_range(start, end, freq="1h")),
            ),
        )

        ds["concno2"].attrs.update(
            data_source=ds["data_source"].values.tolist(),
            var_name=["concno2", "concno2"],
            var_name_input=["concno2", "concno2"],
            model_name=f"CAMS2-83-{model}-day{day}-FC",
        )

        return ds

    start, end = date(2024, 3, 1), date(2024, 3, 5)
    for model, day in product(("ENS", "MOS"), range(4)):
        path = (
            root
            / f"cams2-83/mos-colocated-data/CAMS2-83-{model}-day{day}-FC/concno2_concno2_MOD-CAMS2-83-{model}-day{day}-FC_REF-EEA-UTD_{start:%Y%m%d}_{end:%Y%m%d}_hourly_ALL-wMOUNTAINS.nc"
        )
        path.parent.mkdir(exist_ok=True, parents=True)
        dataset(model, day, start, end).to_netcdf(path)

    start, end = date(2024, 3, 1), date(2024, 3, 2)
    for model in ("ENS", "MOS"):
        path = (
            root
            / f"cams2-83/mos-colocated-data/{model}/concno2_concno2_MOD-{model}_REF-EEA-UTD_{start:%Y%m%d}_{start:%Y%m%d}_hourly_ALL-wMOUNTAINS.nc"
        )
        path.parent.mkdir(exist_ok=True, parents=True)
        ds = dataset(model, 0, start, end)
        ds["concno2"].attrs.update(model_name=model)
        ds.to_netcdf(path)

    return root



from pathlib import Path

import pytest
from pydantic import ValidationError

from pyaerocom import const
from pyaerocom.colocation.colocation_setup import ColocationSetup
from pyaerocom.config import ALL_REGION_NAME

COL_OUT_DEFAULT = Path(const.OUTPUTDIR) / "colocated_data"

default_setup = {
    "model_id": None,
    "obs_id": None,
    "obs_vars": (),
    # "obs_config": None,
    "ts_type": "monthly",
    "start": None,
    "stop": None,
    "filter_name": f"{ALL_REGION_NAME}-wMOUNTAINS",
    "basedir_coldata": COL_OUT_DEFAULT,
    "save_coldata": False,
    "obs_name": None,
    "obs_data_dir": None,
    "obs_use_climatology": False,
    "obs_cache_only": False,
    "obs_vert_type": None,
    "obs_ts_type_read": None,
    "obs_filters": {},
    "model_name": None,
    "model_data_dir": None,
    "model_read_opts": {},
    "read_opts_ungridded": {},
    "model_use_vars": {},
    "model_rename_vars": {},
    "model_add_vars": {},
    "model_to_stp": False,
    "model_ts_type_read": None,
    "model_read_aux": {},
    "model_use_climatology": False,
    "gridded_reader_id": {"model": "ReadGridded", "obs": "ReadGridded"},
    "flex_ts_type": True,
    "min_num_obs": None,
    "resample_how": "mean",
    "obs_remove_outliers": False,
    "model_remove_outliers": False,
    "zeros_to_nan": False,
    "obs_outlier_ranges": {},
    "model_outlier_ranges": {},
    "harmonise_units": False,
    "regrid_res_deg": None,
    "colocate_time": False,
    "reanalyse_existing": True,
    "raise_exceptions": False,
    "keep_data": True,
    "add_meta": {},
}


@pytest.mark.parametrize("stp,should_be", [(ColocationSetup(), default_setup)])
def test_ColocationSetup(stp: ColocationSetup, should_be: dict):
    stp_dict = stp.model_dump()
    for key, val in should_be.items():
        assert key in stp_dict
        if key == "basedir_coldata":
            assert Path(val) == Path(stp_dict["basedir_coldata"])
        else:
            assert val == stp_dict[key], key


def test_ColocationSetup_model_kwargs_validationerror() -> None:
    with pytest.raises(ValidationError, match="Input should be a valid dictionary"):
        ColocationSetup(**default_setup, model_kwargs="not a dict")

    with pytest.raises(ValidationError, match="Key ts_type not allowed in model_kwargs "):
        ColocationSetup(**default_setup, model_kwargs={"emep_vars": {}, "ts_type": "daily"})



from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pyaerocom.io import PyaroConfig, ReadPyaro


def make_csv_test_file(tmp_path: Path) -> Path:
    d = tmp_path / "pyaro"
    file = d / "test_data.csv"

    if not d.exists():
        d.mkdir()
    if file.exists():
        return file

    start = pd.to_datetime("01.01.2010", dayfirst=True)
    end = pd.to_datetime("31.12.2010", dayfirst=True)
    dates = pd.date_range(start, end, freq="D")
    stations = ["NO0002", "GB0881"]
    countries = ["NO", "GB"]
    coords = [(58, 8), (60, -1)]
    species = ["NOx", "SOx", "AOD"]
    area_type = ["Rural", "Urban"]

    with open(file, "w") as f:
        for s in species:
            for i, station in enumerate(stations):
                for j, date in enumerate(dates):
                    delta_t = ["1h", "3D", "2D", "2h"][
                        j % 4
                    ]  # Rotates over the freqs in a deterministic fashion
                    f.write(
                        f"{s}, {station}, {coords[i][1]}, {coords[i][0]}, {np.random.normal(10, 5)}, Gg, {date}, {date+pd.Timedelta(delta_t)},{countries[i]},{area_type[i]} \n"
                    )

    return file


def testconfig(tmp_path: Path) -> PyaroConfig:
    reader_id = "csv_timeseries"

    config1 = PyaroConfig(
        name="test",
        reader_id=reader_id,
        filename_or_obj_or_url=str(make_csv_test_file(tmp_path)),
        filters={},
        name_map={"SOx": "concso4", "AOD": "od550aer"},
    )

    config2 = PyaroConfig(
        name="test2",
        reader_id=reader_id,
        filename_or_obj_or_url=str(make_csv_test_file(tmp_path)),
        filters={},
        name_map={"SOx": "concso4", "AOD": "od550aer"},
    )
    return [config1, config2]


def testconfig_kwargs(tmp_path: Path) -> PyaroConfig:
    reader_id = "csv_timeseries"
    columns = {
        "variable": 0,
        "station": 1,
        "longitude": 2,
        "latitude": 3,
        "value": 4,
        "units": 5,
        "start_time": 6,
        "end_time": 7,
        "altitude": "0",
        "country": 8,
        "standard_deviation": "NaN",
        "flag": "0",
        "area_classification": 9,
    }

    config = PyaroConfig(
        name="test",
        reader_id=reader_id,
        filename_or_obj_or_url=str(make_csv_test_file(tmp_path)),
        filters={},
        name_map={"SOx": "concso4"},
        columns=columns,
    )

    return config


@pytest.fixture
def pyaro_kwargs() -> dict:
    columns = {
        "variable": 0,
        "station": 1,
        "longitude": 2,
        "latitude": 3,
        "value": 4,
        "units": 5,
        "start_time": 6,
        "end_time": 7,
        "altitude": "0",
        "country": 8,
        "standard_deviation": "NaN",
        "flag": "0",
        "area_classification": 9,
    }
    return columns


@pytest.fixture
def pyaro_test_data_file(tmp_path) -> Path:
    return make_csv_test_file(tmp_path)


@pytest.fixture
def pyaro_testconfig(tmp_path) -> PyaroConfig:
    return testconfig(tmp_path=tmp_path)


@pytest.fixture
def pyaro_testconfig_kwargs(tmp_path) -> PyaroConfig:
    return testconfig_kwargs(tmp_path=tmp_path)


@pytest.fixture
def pyaro_testdata(tmp_path) -> ReadPyaro:
    config = testconfig(tmp_path=tmp_path)[0]
    rp = ReadPyaro(config=config)

    return rp


@pytest.fixture
def pyaro_testdata_kwargs(tmp_path) -> ReadPyaro:
    config = testconfig_kwargs(tmp_path=tmp_path)
    rp = ReadPyaro(config=config)

    return rp



import warnings
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np
from cartopy.mpl.geoaxes import GeoAxes
from cartopy.mpl.ticker import LatitudeFormatter, LongitudeFormatter
from geonum.helpers import shifted_color_map
from matplotlib.colors import BoundaryNorm, LogNorm, Normalize
from numpy import ceil, linspace, meshgrid
from pandas import to_datetime

from pyaerocom._warnings import ignore_warnings
from pyaerocom.exceptions import DataDimensionError
from pyaerocom.mathutils import exponent
from pyaerocom.plot.config import COLOR_THEME, MAP_AXES_ASPECT, ColorTheme
from pyaerocom.plot.helpers import (
    calc_figsize,
    calc_pseudolog_cmaplevels,
    custom_mpl,
    projection_from_str,
)
from pyaerocom.region import Region

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning, module="pyaerocom")
    MPL_PARAMS = custom_mpl()


def get_cmap_maps_aerocom(color_theme=None, vmin=None, vmax=None):
    """Get colormap using pyAeroCom color scheme

    Parameters
    ----------
    color_theme : :ColorTheme, optional
        instance of pyaerocom color theme. If None, the default schemes is used
    vmin : float, optional
        lower end of value range (only considered for diverging color maps
        with non-symmetric mapping)
    vmax : float, optional
        upper end of value range only considered for diverging color maps
        with non-symmetric mapping)

    Returns
    -------
    colormap
    """
    warnings.warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )

    if color_theme is None:
        color_theme = COLOR_THEME
    if vmin is not None and vmax is not None and vmin < 0 and vmax > 0:
        cmap = plt.get_cmap(color_theme.cmap_map_div)
        if color_theme.cmap_map_div_shifted:
            cmap = shifted_color_map(vmin, vmax, cmap)
        return cmap
    return plt.get_cmap(color_theme.cmap_map)


def set_map_ticks(ax, xticks=None, yticks=None):
    """Set or update ticks in instance of GeoAxes object (cartopy)

    Parameters
    ----------
    ax : cartopy.GeoAxes
        map axes instance
    xticks : iterable, optional
        ticks of x-axis (longitudes)
    yticks : iterable, optional
        ticks of y-axis (latitudes)

    Returns
    -------
    cartopy.GeoAxes
        modified axes instance
    """
    warnings.warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )
    lonleft, lonright = ax.get_xlim()
    digits = 2 - exponent(lonleft)
    digits = 0 if digits < 0 else digits
    tick_format = f".{digits:d}f"
    if xticks is None:
        num_lonticks = 7 if lonleft == -lonright else 6
        xticks = linspace(lonleft, lonright, num_lonticks)
    if yticks is None:
        latleft, latright = ax.get_ylim()
        num_latticks = 7 if latleft == -latright else 6
        yticks = linspace(latleft, latright, num_latticks)
    ax.set_xticks(xticks, crs=ccrs.PlateCarree())
    ax.set_yticks(yticks, crs=ccrs.PlateCarree())

    lon_formatter = LongitudeFormatter(
        number_format=tick_format, degree_symbol="", dateline_direction_label=True
    )
    lat_formatter = LatitudeFormatter(number_format=tick_format, degree_symbol="")

    ax.xaxis.set_major_formatter(lon_formatter)
    ax.yaxis.set_major_formatter(lat_formatter)

    return ax


def init_map(
    xlim=(-180, 180),
    ylim=(-90, 90),
    figh=8,
    fix_aspect=False,
    xticks=None,
    yticks=None,
    color_theme=COLOR_THEME,
    projection=None,
    title=None,
    gridlines=False,
    fig=None,
    ax=None,
    draw_coastlines=True,
    contains_cbar=False,
):
    """Initalise a map plot

    Parameters
    ----------
    xlim : tuple
        2-element tuple specifying plotted longitude range
    ylim : tuple
        2-element tuple specifying plotted latitude range
    figh : int
        height of figure in inches
    fix_aspect : bool, optional
        if True, the aspect of the GeoAxes instance is kept fix using the
        default aspect ``MAP_AXES_ASPECT`` defined in
        :mod:`pyaerocom.plot.config`
    xticks : iterable, optional
        ticks of x-axis (longitudes)
    yticks : iterable, optional
        ticks of y-axis (latitudes)
    color_theme : ColorTheme
        pyaerocom color theme.
    projection
        projection instance from cartopy.crs module (e.g. PlateCaree). May also
        be string.
    title : str, optional
        title that is supposed to be inserted
    gridlines : bool
        whether or not to add gridlines to the map
    fig : matplotlib.figure.Figure, optional
        instance of matplotlib Figure class. If specified, the former to
        input args (``figh`` and ``fix_aspect``) are ignored. Note that the
        Figure is wiped clean before plotting, so any plotted content will be
        lost
    ax : GeoAxes, optional
        axes in which the map is plotted
    draw_coastlines : bool
        whether or not to draw coastlines
    contains_cbar : bool
        whether or not a colorbar is intended to be added to the figure (
        impacts the aspect ratio of the figure).

    Returns
    -------
    ax : cartopy.mpl.geoaxes.GeoAxes
        axes instance
    """
    warnings.warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )
    if projection is None:
        projection = ccrs.PlateCarree()
    elif isinstance(projection, str):
        projection = projection_from_str(projection)
    elif not isinstance(projection, ccrs.Projection):
        raise ValueError("Input for projection needs to be instance of cartopy.crs.Projection")

    if not isinstance(ax, GeoAxes):
        if fig is None:
            if not fix_aspect:
                figsize = calc_figsize(xlim, ylim)
            else:
                figw = figh * fix_aspect
                figsize = (figw, figh)

            fig = plt.figure(figsize=figsize)
        else:
            fig.clf()
        if contains_cbar:
            ax = fig.add_axes([0.1, 0.12, 0.75, 0.8], projection=projection)
        else:
            ax = fig.add_axes([0.1, 0.12, 0.85, 0.8], projection=projection)

    if fix_aspect:
        ax.set_aspect(MAP_AXES_ASPECT)

    if not isinstance(color_theme, ColorTheme):
        if isinstance(color_theme, str):
            color_theme = ColorTheme(color_theme)
        else:
            color_theme = COLOR_THEME

    if draw_coastlines:
        ax.coastlines(color=color_theme.color_coastline)

    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    ax = set_map_ticks(ax, xticks, yticks)
    ax.set_xlabel("Longitude")
    ax.set_ylabel("Latitude")

    if title is not None:
        ax.set_title(title)
    if gridlines:
        ax.gridlines()

    return ax


def plot_griddeddata_on_map(
    data,
    lons=None,
    lats=None,
    var_name=None,
    unit=None,
    xlim=(-180, 180),
    ylim=(-90, 90),
    vmin=None,
    vmax=None,
    add_zero=False,
    c_under=None,
    c_over=None,
    log_scale=True,
    discrete_norm=True,
    cbar_levels=None,
    cbar_ticks=None,
    add_cbar=True,
    cmap=None,
    cbar_ticks_sci=False,
    color_theme=None,
    ax=None,
    ax_cbar=None,
    **kwargs,
):
    """Make a plot of gridded data onto a map

    Parameters
    ----------
    data : ndarray
        2D data array
    lons : ndarray
        longitudes of data
    lats : ndarray
        latitudes of data
    var_name : :obj:`str`, optional
        name of variable that is plotted
    xlim : tuple
        2-element tuple specifying plotted longitude range
    ylim : tuple
        2-element tuple specifying plotted latitude range
    vmin : :obj:`float`, optional
        lower value of colorbar range
    vmax : :obj:`float`, optional
        upper value of colorbar range
    add_zero : bool
        if True and vmin is not 0, then, the colorbar is extended down to 0.
        This may be used, e.g. for logarithmic scales that should include 0.
    c_under : :obj:`float`, optional
        colour of data values smaller than ``vmin``
    c_over : :obj:`float`, optional
        colour of data values exceeding ``vmax``
    log_scale : bool
        if True, the value to color mapping is done in a pseudo log scale
        (see :func:`get_cmap_levels_auto` for implementation)
    discrete_norm : bool
        if True, color mapping will be subdivided into discrete intervals
    cbar_levels : iterable, optional
        discrete colorbar levels. Will be computed automatically, if None
        (and applicable)
    cbar_ticks : iterable, optional
        ticks of colorbar levels. Will be computed automatically, if None
        (and applicable)

    Returns
    -------
    fig
        matplotlib figure instance containing plot result. Use
        ``fig.axes[0]`` to access the map axes instance (e.g. to modify the
        title or lon / lat range, etc.)
    """
    warnings.warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )
    if color_theme is None:
        color_theme = COLOR_THEME
    if add_cbar:
        kwargs["contains_cbar"] = True
    if ax is None:
        ax = init_map(xlim, ylim, color_theme=color_theme, **kwargs)
    if not isinstance(ax, GeoAxes):
        raise ValueError("Invalid input for ax, need GeoAxes")
    fig = ax.figure

    from pyaerocom.griddeddata import GriddedData

    if not isinstance(data, GriddedData):
        raise ValueError("need GriddedData")

    if not data.has_latlon_dims:
        raise DataDimensionError("Input data needs to have latitude and longitude dimension")
    if not data.ndim == 2:
        if not data.ndim == 3 or "time" not in data.dimcoord_names:
            raise DataDimensionError(
                "Input data needs to be 2 dimensional "
                "or 3D with time being the 3rd "
                "dimension"
            )
        data.reorder_dimensions_tseries()
        data = data[0]

    lons = data.longitude.points
    lats = data.latitude.points
    data = data.grid.data

    if add_cbar and ax_cbar is None:
        ax_cbar = _add_cbar_axes(ax)

    X, Y = meshgrid(lons, lats)

    bounds = None
    if cbar_levels is not None:  # user provided levels of colorbar explicitely
        if vmin is not None or vmax is not None:
            raise ValueError("Please provide either vmin/vmax OR cbar_levels")
        bounds = list(cbar_levels)
        low, high = bounds[0], bounds[-1]
        if add_zero and low > 0:
            bounds.insert(0, 0)  # insert zero bound
        if cmap is None:
            cmap = get_cmap_maps_aerocom(color_theme, low, high)
        elif isinstance(cmap, str):
            cmap = plt.get_cmap(cmap)
        norm = BoundaryNorm(boundaries=bounds, ncolors=cmap.N, clip=False)
    else:
        with ignore_warnings(RuntimeWarning, "All-NaN axis encountered"):
            dmin = np.nanmin(data)
            dmax = np.nanmax(data)

        if any([np.isnan(x) for x in [dmin, dmax]]):
            raise ValueError("Cannot plot map of data: all values are NaN")
        elif dmin == dmax:
            raise ValueError(f"Minimum value in data equals maximum value: {dmin}")
        if vmin is None:
            vmin = dmin
        else:
            if vmin < 0 and log_scale:
                log_scale = False
        if vmax is None:
            vmax = dmax
        if exponent(vmin) == exponent(vmax):
            log_scale = False
        if log_scale:  # no negative values allowed
            if vmin < 0:
                vmin = data[data > 0].min()
                if (
                    c_under is None
                ):  # special case, set c_under to indicate that there is values below 0
                    c_under = "r"
            if cmap is None:
                cmap = get_cmap_maps_aerocom(color_theme, vmin, vmax)
            elif isinstance(cmap, str):
                cmap = plt.get_cmap(cmap)
            if discrete_norm:
                # to compute upper range of colour range, round up vmax
                exp = float(exponent(vmax) - 1)
                vmax_colors = ceil(vmax / 10**exp) * 10**exp
                bounds = calc_pseudolog_cmaplevels(vmin=vmin, vmax=vmax_colors, add_zero=add_zero)
                norm = BoundaryNorm(boundaries=bounds, ncolors=cmap.N, clip=False)

            else:
                norm = LogNorm(vmin=vmin, vmax=vmax, clip=True)
        else:
            if add_zero and vmin > 0:
                vmin = 0
            if cmap is None:
                cmap = get_cmap_maps_aerocom(color_theme, vmin, vmax)
            elif isinstance(cmap, str):
                cmap = plt.get_cmap(cmap)
            if discrete_norm:
                bounds = np.linspace(vmin, vmax, 10)
                norm = BoundaryNorm(boundaries=bounds, ncolors=cmap.N, clip=False)
            else:
                norm = Normalize(vmin=vmin, vmax=vmax)
    cbar_extend = "neither"
    if c_under is not None:
        cmap = cmap.copy()
        cmap.set_under(c_under)
        cbar_extend = "min"
        if bounds is not None:
            bounds.insert(0, bounds[0] - bounds[1])
    if c_over is not None:
        cmap = cmap.copy()
        cmap.set_over(c_over)
        if bounds is not None:
            bounds.append(bounds[-1] + bounds[-2])
        if cbar_extend == "min":
            cbar_extend = "both"
        else:
            cbar_extend = "max"
    fig.norm = norm
    disp = ax.pcolormesh(X, Y, data, cmap=cmap, norm=norm, shading="auto")

    if add_cbar:
        cbar = fig.colorbar(disp, extend=cbar_extend, cax=ax_cbar, shrink=0.8)
        fig.cbar = cbar

        if var_name is not None:
            var_str = var_name  # + VARS.unit_str
            if unit is not None:
                if str(unit) not in ["1", "no_unit"]:
                    var_str += f" [{unit}]"

            cbar.set_label(var_str)

        if cbar_ticks:
            cbar.set_ticks(cbar_ticks)
        if cbar_ticks_sci:
            lbls = []
            for lbl in cbar.ax.get_yticklabels():
                tstr = lbl.get_text()
                if bool(tstr):
                    lbls.append(f"{float(tstr):.1e}")
                else:
                    lbls.append("")
            cbar.ax.set_yticklabels(lbls)

    return fig


def _add_cbar_axes(ax):  # , where='right'):
    _loc = ax.bbox._bbox
    fig = ax.figure

    ax_cbar = fig.add_axes([_loc.x1 + 0.02, _loc.y0, 0.02, _loc.y1 - _loc.y0])
    return ax_cbar


def plot_map_aerocom(data, region, **kwargs):
    """High level map plotting function for Aerocom default plotting

    Note
    ----
    This function does not iterate over a cube in time, but uses the
    first available time index in the data.

    Parameters
    ----------
    data : GriddedData
        input data from one timestamp (if data contains more than one time
        stamp, the first index is used)
    region : str or Region
        valid region ID or region

    """
    warnings.warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )
    from pyaerocom import GriddedData

    if not isinstance(data, GriddedData):
        raise ValueError(
            "This plotting method needs an instance of pyaerocom "
            f"GriddedData on input, got: {type(data)}"
        )

    if isinstance(region, str):
        region = Region(region)
    elif not isinstance(region, Region):
        raise ValueError("Invalid input for region, need None, str or Region")

    data = data.filter_region(region.region_id)
    s = data.plot_settings

    vmin, vmax, levs = s.map_vmin, s.map_vmax, None
    if isinstance(s.map_cbar_levels, list) and len(s.map_cbar_levels) > 0:
        vmin, vmax = None, None
        levs = s.map_cbar_levels
    fig = plot_griddeddata_on_map(
        data,
        xlim=region.lon_range_plot,
        ylim=region.lat_range_plot,
        vmin=vmin,
        vmax=vmax,
        c_over=s.map_c_over,
        c_under=s.map_c_under,
        cbar_levels=levs,
        xticks=region.lon_ticks,
        yticks=region.lat_ticks,
        cbar_ticks=s.map_cbar_ticks,
        **kwargs,
    )
    ax = fig.axes[0]

    # annotate model in lower left corner
    lonr, latr = region.lon_range_plot, region.lat_range_plot
    ax.annotate(
        data.data_id,
        xy=(lonr[0] + (lonr[1] - lonr[0]) * 0.03, latr[0] + (latr[1] - latr[0]) * 0.03),
        xycoords="data",
        horizontalalignment="left",
        color="black",
        fontsize=MPL_PARAMS["axes.titlesize"] + 2,
        bbox=dict(boxstyle="square", facecolor="white", edgecolor="none", alpha=0.7),
    )
    ax.annotate(
        "source: AEROCOM",
        xy=(0.97, 0.03),
        xycoords="figure fraction",
        horizontalalignment="right",
        fontsize=MPL_PARAMS["xtick.labelsize"],
        bbox=dict(boxstyle="square", facecolor="none", edgecolor="black"),
    )

    var = data.var_name.upper()
    avg = data.mean()
    start = to_datetime(data.start).strftime("%Y%m%d")
    tit = f"{var} {start} mean {avg.round(3)}"
    ax.set_title(tit)
    return fig


def plot_nmb_map_colocateddata(
    coldata,
    in_percent=True,
    vmin=-100,
    vmax=100,
    cmap=None,
    s=80,
    marker=None,
    step_bounds=None,
    add_cbar=True,
    norm=None,
    cbar_extend=None,
    add_mean_edgecolor=True,
    ax=None,
    ax_cbar=None,
    cbar_outline_visible=False,
    cbar_orientation=None,
    ref_label=None,
    stats_area_weighted=False,
    **kwargs,
):
    """Plot map of normalised mean bias from instance of ColocatedData

    Parameters
    ----------
    coldata : ColocatedData
        data object
    in_percent : bool
        plot bias in percent
    vmin : int
        minimum value of colormapping
    vmax : int
        maximum value of colormapping
    cmap : str or cmap
        colormap used, defaults to bwr
    s : int
        size of marker
    marker : str
        marker used
    step_bounds : int, optional
        step used for discrete colormapping (if None, continuous is used)
    cbar_extend : str
        extend colorbar
    ax : GeoAxes, optional
        axes into which the bias is supposed to be plotted
    ax_cbar : plt.Axes, optional
        axes for colorbar
    cbar_outline_visible : bool
        if False, borders of colorbar are removed
    cbar_orientation : str
        e.g. 'vertical', defaults to 'vertical'
    **kwargs
        keyword args passed to :func:`init_map`

    Returns
    -------
    GeoAxes
    """
    warnings.warn(
        "matplotlib based plotting is no longer directly supported. This function may be removed in future versions.",
        DeprecationWarning,
    )
    if cbar_extend is None:
        cbar_extend = "both"
    if cbar_orientation is None:
        cbar_orientation = "vertical"
    if cmap is None:
        cmap = "bwr"
    if marker is None:
        marker = "o"
    try:
        mec = kwargs.pop("mec")
    except KeyError:
        try:
            mec = kwargs.pop("markeredgecolor")
        except KeyError:
            mec = "face"

    try:
        mew = kwargs.pop("mew")
    except KeyError:
        mew = 1
    if coldata.ndim not in (3, 4):
        raise DataDimensionError("only 3D or 4D colocated data objects are supported")
    assert "time" in coldata.dims

    mean_bias = coldata.calc_nmb_array()

    if mean_bias.ndim == 1:
        (lats, lons, data) = mean_bias.latitude, mean_bias.longitude, mean_bias.data
    elif "latitude" in mean_bias.dims and "longitude" in mean_bias.dims:
        stacked = mean_bias.stack(latlon=["latitude", "longitude"])
        valid = ~stacked.isnull()
        coords = stacked.latlon[valid].values
        lats, lons = list(zip(*list(coords)))
        data = stacked.data[tuple(valid)]

    if ref_label is None:
        ref_label = coldata.metadata["data_source"][0]

    if in_percent:
        data *= 100
    if ax is None:
        ax = init_map(contains_cbar=True, **kwargs)

    if not isinstance(ax, GeoAxes):
        raise TypeError("Input axes need to be instance of cartopy.GeoAxes")

    fig = ax.figure

    if isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)
    if norm is None and step_bounds is not None:
        bounds = np.arange(vmin, vmax + step_bounds, step_bounds)
        norm = BoundaryNorm(boundaries=bounds, ncolors=cmap.N, clip=False)

    if add_mean_edgecolor:
        nn = Normalize(vmin=vmin, vmax=vmax)
        nmb = coldata.calc_statistics(use_area_weights=stats_area_weighted)["nmb"]
        if in_percent:
            nmb *= 100
        ec = cmap(nn(nmb))
    else:
        ec = mec
    _sc = ax.scatter(
        lons,
        lats,
        c=data,
        marker=marker,
        cmap=cmap,
        vmin=vmin,
        vmax=vmax,
        s=s,
        norm=norm,
        label=ref_label,
        edgecolors=ec,
        linewidths=mew,
    )
    if add_cbar:
        if ax_cbar is None:
            ax_cbar = _add_cbar_axes(ax)
        cbar = fig.colorbar(_sc, extend=cbar_extend, cax=ax_cbar, orientation=cbar_orientation)

        cbar.outline.set_visible(cbar_outline_visible)
        cbar.set_label("NMB [%]")

    return ax



from __future__ import annotations

from pathlib import Path

import pytest

from pyaerocom import const
from pyaerocom.io.icpforests.metadata import MetadataReader as ReadICPForestMeta
from pyaerocom.io.icpforests.reader import ReadICPForest
from tests.fixtures.data_access import TEST_DATA, DataForTests

# station names are not consistent between variables!
# Station names are kind of limited since we took only the last 5000 lines from the data file for the test data set
STATION_NAMES = ("BG-1-1", "LV-15-2", "BG-4-3")

VARS_DEFAULT = {"wetoxn"}
VARS_PROVIDED = VARS_DEFAULT

ICP_PATH = DataForTests(TEST_DATA["ICPFORESTS.Subset"].relpath).path


@pytest.fixture(scope="module")
def reader():
    return ReadICPForest("ICPFORESTS.Subset")


@pytest.fixture(scope="module")
def meta_reader() -> ReadICPForestMeta:
    return ReadICPForestMeta(str(ICP_PATH))


@pytest.fixture()
def station_files(station: str) -> list[Path]:
    p = ICP_PATH.glob("dp_dem.csv")
    files = [x for x in p if x.is_file()]
    # assert files, f"no files for {station}"
    assert files
    return files


def test_DATASET_NAME(reader: ReadICPForest):
    assert reader.DATA_ID == const.ICPFORESTS_NAME


def test_DEFAULT_VARS(reader: ReadICPForest):
    assert set(reader.DEFAULT_VARS) >= VARS_DEFAULT


def test_METADATA(meta_reader: ReadICPForestMeta):
    assert len(meta_reader.deposition_type) >= 3, "found less deposition types than expected"


def test_PROVIDES_VARIABLES(reader: ReadICPForest):
    assert set(reader.PROVIDES_VARIABLES) >= VARS_PROVIDED


def test_read_file(
    reader: ReadICPForest,
):
    data = reader.read(vars_to_retrieve=VARS_DEFAULT)
    assert set(data.contains_vars) == VARS_DEFAULT


def test_read_station(
    reader: ReadICPForest,
):
    data = reader.read(
        vars_to_retrieve=VARS_DEFAULT,
    )
    for station in STATION_NAMES:
        assert station in data.unique_station_names


def test_reader_gives_correct_ICP_PATH(reader: ReadICPForest):
    assert str(ICP_PATH) == reader.data_dir



import logging
import os

import numpy as np
import pandas as pd
import xarray

from pyaerocom import const
from pyaerocom.exceptions import DataUnitError
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData
from pyaerocom.units_helpers import get_unit_conversion_fac
from pyaerocom.variable import Variable
from pyaerocom.vertical_profile import VerticalProfile

logger = logging.getLogger(__name__)


class ReadEarlinet(ReadUngriddedBase):
    """Interface for reading of EARLINET data"""

    #: Mask for identifying datafiles
    _FILEMASK = "*.*"

    #: version log of this class (for caching)
    __version__ = "0.18_" + ReadUngriddedBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.EARLINET_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.EARLINET_NAME]

    #: default variables for read method
    DEFAULT_VARS = ["bsc532aer", "ec532aer"]

    # These are applied to all files. If new cloud filter names are discovered they should be added here.
    # As of 20.08.24, however, there are still files with less reliable data which get through the filters.
    # https://github.com/metno/pyaerocom/issues/1310
    CLOUD_FILTERS = {
        "cloud_mask_type": 0,  # "no_cloudmask_available manual_cloudmask automatic_cloudmask"
        "cirrus_contamination": 2,  # "not_available no_cirrus cirrus_detected"
    }

    #: all data values that exceed this number will be set to NaN on read. This
    #: is because iris, xarray, etc. assign a FILL VALUE of the order of e36
    #: to missing data in the netcdf files
    _MAX_VAL_NAN = 1e6

    #: variable name of altitude in files
    ALTITUDE_ID = "altitude"

    #: temporal resolution
    # Note: This is an approximation based on the fact that MOST of the data appears to be collected
    # at an hourly reoslution. Some files are a little less, but typically this is the case
    TS_TYPE = "hourly"

    VAR_PATTERNS_FILE = {
        "ec532aer": "_e0532",
        "ec355aer": "_e0355",
        "bsc532aer": "_b0532",
        "bsc355aer": "_b0355",
        "bsc1064aer": "_b1064",
        # "zdust": "*.e*", # not sure if EARLINET has this anymore
    }

    #: dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    VAR_NAMES_FILE = {
        "ec532aer": "extinction",
        "ec355aer": "extinction",
        "ec1064aer": "extinction",
        "bsc532aer": "backscatter",
        "bsc355aer": "backscatter",
        "bsc1064aer": "backscatter",
        "zdust": "DustLayerHeight",  # not sure if EARLINET has this anymore
    }

    META_NAMES_FILE = dict(
        location="location",
        start_utc="measurement_start_datetime",
        stop_utc="measurement_stop_datetime",
        # wavelength_det="DetectionWavelength_nm",
        # res_raw_m="ResolutionRaw_meter",
        instrument_name="system",
        comment="comment",
        PI="PI",
        dataset_name="title",
        # station_name="station_ID",
        website="references",
        wavelength_emis="wavelength",
        # detection_mode="DetectionMode",
        # res_eval="ResolutionEvaluated",
        # input_params="InputParameters",
        altitude="altitude",
        # eval_method="backscatter_evaluation_method",
    )
    #: metadata keys that are needed for reading (must be values in
    #: :attr:`META_NAMES_FILE`)
    META_NEEDED = [
        "location",
        "measurement_start_datetime",
        "measurement_start_datetime",
    ]

    #: Metadata keys from :attr:`META_NAMES_FILE` that are additional to
    #: standard keys defined in :class:`StationMetaData` and that are supposed
    #: to be inserted into :class:`UngriddedData` object created in :func:`read`
    KEEP_ADD_META = [
        "location",
        "wavelength",
        "zenith_angle",
        "comment",
        "shots",
        "backscatter_evaluation_method",
    ]

    #: Attribute access names for unit reading of variable data
    VAR_UNIT_NAMES = dict(
        extinction=["units"],
        backscatter=["units"],
        dustlayerheight=["units"],
        altitude="units",
    )
    #: Variable names of uncertainty data
    ERR_VARNAMES = dict(
        ec532aer="error_extinction",
        ec355aer="error_extinction",
    )

    #: If true, the uncertainties are also read (where available, cf. ERR_VARNAMES)
    READ_ERR = True

    PROVIDES_VARIABLES = list(VAR_PATTERNS_FILE)

    EXCLUDE_CASES = ["cirrus.txt"]

    def __init__(self, data_id=None, data_dir=None):
        # initiate base class
        super().__init__(data_id=data_id, data_dir=data_dir)
        # make sure everything is properly set up
        if not all(
            [x in self.VAR_PATTERNS_FILE for x in self.PROVIDES_VARIABLES]
        ):  # pragma: no cover
            raise AttributeError(
                "Please specify file search masks in "
                "header dict VAR_PATTERNS_FILE for each "
                "variable defined in PROVIDES_VARIABLES"
            )
        elif not all(
            [x in self.VAR_NAMES_FILE for x in self.PROVIDES_VARIABLES]
        ):  # pragma: no cover
            raise AttributeError(
                "Please specify file search masks in "
                "header dict VAR_NAMES_FILE for each "
                "variable defined in PROVIDES_VARIABLES"
            )
        #: private dictionary containing loaded Variable instances,
        self._var_info = {}

        #: files that are supposed to be excluded from reading
        self.exclude_files = []

        #: files that were actually excluded from reading
        self.excluded_files = []

        self.is_vertical_profile = True

    def read_file(self, filename, vars_to_retrieve=None, read_err=None, remove_outliers=True):
        """Read EARLINET file and return it as instance of :class:`StationData`

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        read_err : bool
            if True, uncertainty data is also read (where available).
        remove_outliers : bool
            if True, outliers are removed for each variable using the
            `minimum` and `maximum` attributes for that variable (accessed
            via pyaerocom.const.VARS[var_name]).

        Returns
        -------
        StationData
            dict-like object containing results
        """
        if read_err is None:  # use default setting
            read_err = self.READ_ERR
        if isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        _vars = []
        for var in vars_to_retrieve:
            if (
                var in self.VAR_PATTERNS_FILE
            ):  # make sure to only read what is supported by this file
                if self.VAR_PATTERNS_FILE[var] in filename:
                    _vars.append(var)
            elif var in self.AUX_REQUIRES:
                _vars.append(var)
            else:
                raise ValueError(f"{var} is not supported")

        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(_vars)

        # create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        data_out["station_id"] = filename.split("/")[-1].split("_")[
            2
        ]  # loss of generality but should work. can also get from reading file if needed: data_in.station_ID
        data_out["data_id"] = self.data_id
        data_out["ts_type"] = self.TS_TYPE

        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            if var not in self._var_info:
                self._var_info[var] = Variable(var)
        var_info = self._var_info

        # Iterate over the lines of the file
        self.logger.debug(f"Reading file {filename}")

        with xarray.open_dataset(filename, engine="netcdf4") as data_in:
            for filter in self.CLOUD_FILTERS:
                if filter in data_in.variables:
                    if data_in.variables[filter].item() == self.CLOUD_FILTERS[filter]:
                        self.logger.debug(f"Skipping {filename} due to cloud filtering")
                        continue

            # getting the coords since no longer in metadata
            # Put also just in the attributes. not sure why appears twice
            data_out["station_coords"]["longitude"] = data_out["longitude"] = np.float64(
                data_in["longitude"].values
            )
            data_out["station_coords"]["latitude"] = data_out["latitude"] = np.float64(
                data_in["latitude"].values
            )
            data_out["altitude"] = np.float64(
                data_in[
                    "altitude"
                ].values  # altitude is defined in EARLINET in terms of altitude above sea level
            )  # Note altitude is an array for the data, station altitude is different
            data_out["station_coords"]["altitude"] = np.float64(data_in.station_altitude)
            data_out["altitude_attrs"] = data_in[
                "altitude"
            ].attrs  # get attrs for altitude units + extra

            # get intersection of metadaa in ddataa_out and data_in
            for k, v in self.META_NAMES_FILE.items():
                if v in self.META_NEEDED:
                    _meta = data_in.attrs[v]
                else:
                    try:
                        _meta = data_in.attrs[v]
                    except Exception:  # pragma: no cover
                        _meta = None
                data_out[k] = _meta

            # get metadata expected in StationData but not in data_in's metadata
            data_out["wavelength_emis"] = data_in["wavelength"]
            data_out["shots"] = np.float64(data_in["shots"])
            data_out["zenith_angle"] = np.float64(data_in["zenith_angle"])
            data_out["filename"] = filename
            if "Lev02" in filename:
                data_out["data_level"] = 2
            loc_split = data_in.attrs["location"].split(", ")
            data_out["station_name"] = loc_split[0]
            if len(loc_split) > 1:
                data_out["country"] = loc_split[1]

            dtime = (
                pd.Timestamp(data_in.measurement_start_datetime).to_numpy().astype("datetime64[s]")
            )
            stop = (
                pd.Timestamp(data_in.measurement_stop_datetime).to_numpy().astype("datetime64[s]")
            )

            # in case measurement goes over midnight into a new day
            if stop < dtime:
                stop = stop + np.timedelta64(1, "[D]")

            data_out["dtime"] = [dtime]
            data_out["stopdtime"] = [stop]
            data_out["has_zdust"] = False

            for var in vars_to_read:
                data_out["var_info"][var] = {}
                err_read = False
                unit_ok = False
                outliers_removed = False
                has_altitude = False

                netcdf_var_name = self.VAR_NAMES_FILE[var]
                # check if the desired variable is in the file
                if netcdf_var_name not in data_in.variables:
                    self.logger.warning(f"Variable {var} not found in file {filename}")
                    continue

                info = var_info[var]
                # xarray.DataArray
                arr = data_in.variables[netcdf_var_name]
                # the actual data as numpy array (or float if 0-D data, e.g. zdust)
                val = np.squeeze(np.float64(arr))  # squeeze to 1D array

                # CONVERT UNIT
                unit = None

                unames = self.VAR_UNIT_NAMES[netcdf_var_name]
                for u in unames:
                    if u in arr.attrs:
                        unit = arr.attrs[u]
                if unit is None:
                    raise DataUnitError(f"Unit of {var} could not be accessed in file {filename}")
                unit_fac = None
                try:
                    to_unit = self._var_info[var].units
                    unit_fac = get_unit_conversion_fac(unit, to_unit)
                    val *= unit_fac
                    unit = to_unit
                    unit_ok = True
                except Exception as e:
                    logger.warning(
                        f"Failed to convert unit of {var} in file {filename} (Earlinet): "
                        f"Error: {repr(e)}"
                    )

                # import errors if applicable
                err = np.nan
                if read_err and var in self.ERR_VARNAMES:
                    err_name = self.ERR_VARNAMES[var]
                    if err_name in data_in.variables:
                        err = np.squeeze(np.float64(data_in.variables[err_name]))
                        if unit_ok:
                            err *= unit_fac
                        err_read = True

                # 1D variable
                if var == "zdust":
                    if not val.ndim == 0:
                        raise ValueError("Fatal: dust layer height data must be single value")

                    if unit_ok and info.minimum < val < info.maximum:
                        logger.warning(f"zdust value {val} out of range, setting to NaN")
                        val = np.nan

                    if np.isnan(val):
                        self.logger.warning(
                            f"Invalid value of variable zdust in file {filename}. Skipping...!"
                        )
                        continue

                    data_out["has_zdust"] = True
                    data_out[var] = val

                else:
                    if not val.ndim == 1:
                        raise ValueError("Extinction data must be one dimensional")
                    elif len(val) == 0:
                        continue  # no data
                    # Remove NaN equivalent values
                    val[val > self._MAX_VAL_NAN] = np.nan

                    wvlg = var_info[var].wavelength_nm
                    wvlg_str = self.META_NAMES_FILE["wavelength_emis"]

                    if not wvlg == float(data_in[wvlg_str]):
                        self.logger.info("No wavelength match")
                        continue

                    alt_id = self.ALTITUDE_ID
                    alt_data = data_in.variables[alt_id]

                    alt_vals = np.float64(alt_data)
                    alt_unit = alt_data.attrs[self.VAR_UNIT_NAMES[alt_id]]
                    to_alt_unit = const.VARS["alt"].units
                    if not alt_unit == to_alt_unit:
                        try:
                            alt_unit_fac = get_unit_conversion_fac(alt_unit, to_alt_unit)
                            alt_vals *= alt_unit_fac
                            alt_unit = to_alt_unit
                        except Exception as e:
                            self.logger.warning(f"Failed to convert unit: {repr(e)}")
                    has_altitude = True

                    # remove outliers from data, if applicable
                    if remove_outliers and unit_ok:
                        # REMOVE OUTLIERS
                        outlier_mask = np.logical_or(val < info.minimum, val > info.maximum)
                        val[outlier_mask] = np.nan

                        if err_read:
                            err[outlier_mask] = np.nan
                        outliers_removed = True
                    # remove outliers from errors if applicable
                    if err_read:
                        err[err > self._MAX_VAL_NAN] = np.nan

                    # create instance of ProfileData
                    profile = VerticalProfile(
                        data=val,
                        altitude=alt_vals,
                        dtime=dtime,
                        var_name=var,
                        data_err=err,
                        var_unit=unit,
                        altitude_unit=alt_unit,
                    )

                    # Write everything into profile
                    data_out[var] = profile

                data_out["var_info"][var].update(
                    unit_ok=unit_ok,
                    err_read=err_read,
                    outliers_removed=outliers_removed,
                    has_altitute=has_altitude,
                )
        return data_out

    def read(
        self,
        vars_to_retrieve=None,
        files=None,
        first_file=None,
        last_file=None,
        read_err=None,
        remove_outliers=True,
        pattern=None,
    ):
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.
        first_file : :obj:`int`, optional
            index of first file in file list to read. If None, the very first
            file in the list is used
        last_file : :obj:`int`, optional
            index of last file in list to read. If None, the very last file
            in the list is used
        read_err : bool
            if True, uncertainty data is also read (where available). If
            unspecified (None), then the default is used (cf. :attr:`READ_ERR`)
         pattern : str, optional
            string pattern for file search (cf :func:`get_file_list`)

        Returns
        -------
        UngriddedData
            data object
        """

        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        if read_err is None:
            read_err = self.READ_ERR

        if files is None:
            if len(self.files) == 0:
                self.get_file_list(vars_to_retrieve, pattern=pattern)
            files = self.files

        # turn files into a list becauase I suspect there may be a bug if you don't do this
        if isinstance(files, str):
            files = [files]

        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)

        files = files[
            first_file : last_file + 1
        ]  # think need to +1 here in order to actually get desired subset

        self.read_failed = []

        data_obj = UngriddedData()
        data_obj.is_vertical_profile = True
        col_idx = data_obj.index
        meta_key = -1.0
        idx = 0

        # assign metadata object
        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx

        # last_station_id = ''
        num_files = len(files)

        disp_each = int(num_files * 0.1)
        if disp_each < 1:
            disp_each = 1

        VAR_IDX = -1
        for i, _file in enumerate(files):
            if i % disp_each == 0:
                print(f"Reading file {i + 1} of {num_files} ({type(self).__name__})")
            try:
                stat = self.read_file(
                    _file,
                    vars_to_retrieve=vars_to_retrieve,
                    read_err=read_err,
                    remove_outliers=remove_outliers,
                )
                if not any([var in stat.vars_available for var in vars_to_retrieve]):
                    self.logger.info(
                        f"Station {stat.station_name} contains none of the desired variables. Skipping station..."
                    )
                    continue
                # if last_station_id != station_id:
                meta_key += 1
                # Fill the metatdata dict
                # the location in the data set is time step dependant!
                # use the lat location here since we have to choose one location
                # in the time series plot
                metadata[meta_key] = {}
                metadata[meta_key].update(stat.get_meta())
                for add_meta in self.KEEP_ADD_META:
                    if add_meta in stat:
                        metadata[meta_key][add_meta] = stat[add_meta]
                # metadata[meta_key]['station_id'] = station_id

                metadata[meta_key]["data_revision"] = self.data_revision
                metadata[meta_key]["variables"] = []
                metadata[meta_key]["var_info"] = {}
                # this is a list with indices of this station for each variable
                # not sure yet, if we really need that or if it speeds up things
                meta_idx[meta_key] = {}
                # last_station_id = station_id

                # Is floating point single value
                time = stat.dtime[0]
                for var in stat.vars_available:
                    if var not in data_obj.var_idx:
                        VAR_IDX += 1
                        data_obj.var_idx[var] = VAR_IDX

                    var_idx = data_obj.var_idx[var]

                    val = stat[var]
                    metadata[meta_key]["var_info"][var] = vi = {}
                    if isinstance(val, VerticalProfile):
                        altitude = val.altitude
                        data = val.data
                        add = len(data)
                        err = val.data_err
                        metadata[meta_key]["var_info"]["altitude"] = via = {}

                        vi.update(val.var_info[var])
                        via.update(val.var_info["altitude"])
                    else:
                        add = 1
                        altitude = np.nan
                        data = val
                        if var in stat.data_err:
                            err = stat.err[var]
                        else:
                            err = np.nan
                    vi.update(stat.var_info[var])
                    stop = idx + add
                    # check if size of data object needs to be extended
                    if stop >= data_obj._ROWNO:
                        # if totnum < data_obj._CHUNKSIZE, then the latter is used
                        data_obj.add_chunk(add)

                    # write common meta info for this station
                    data_obj._data[idx:stop, col_idx["latitude"]] = stat["station_coords"][
                        "latitude"
                    ]
                    data_obj._data[idx:stop, col_idx["longitude"]] = stat["station_coords"][
                        "longitude"
                    ]
                    data_obj._data[idx:stop, col_idx["altitude"]] = stat["station_coords"][
                        "altitude"
                    ]
                    data_obj._data[idx:stop, col_idx["meta"]] = meta_key

                    # write data to data object
                    data_obj._data[idx:stop, col_idx["time"]] = time
                    data_obj._data[idx:stop, col_idx["stoptime"]] = stat.stopdtime[0]
                    data_obj._data[idx:stop, col_idx["data"]] = data
                    data_obj._data[idx:stop, col_idx["dataaltitude"]] = altitude
                    data_obj._data[idx:stop, col_idx["varidx"]] = var_idx

                    if read_err:
                        data_obj._data[idx:stop, col_idx["dataerr"]] = err

                    if var not in meta_idx[meta_key]:
                        meta_idx[meta_key][var] = []
                    meta_idx[meta_key][var].extend(list(range(idx, stop)))

                    if var not in metadata[meta_key]["variables"]:
                        metadata[meta_key]["variables"].append(var)

                    idx += add

            except Exception as e:
                self.read_failed.append(_file)
                self.logger.exception(
                    f"Failed to read file {os.path.basename(_file)} (ERR: {repr(e)})"
                )

        # shorten data_obj._data to the right number of points
        data_obj._data = data_obj._data[:idx]

        return data_obj

    def _get_exclude_filelist(self):  # pragma: no cover
        """Get list of filenames that are supposed to be ignored"""
        exclude = []
        import glob

        files = glob.glob(f"{self.data_dir}/EXCLUDE/*.txt")
        for i, file in enumerate(files):
            if os.path.basename(file) not in self.EXCLUDE_CASES:
                continue
            count = 0
            num = None
            indata = False
            with open(file) as f:
                for line in f:
                    if indata:
                        exclude.append(line.strip())
                        count += 1
                    elif "Number of" in line:
                        num = int(line.split(":")[1].strip())
                        indata = True

            if not count == num:
                raise Exception
        self.exclude_files = list(dict.fromkeys(exclude))
        return self.exclude_files

    def get_file_list(self, vars_to_retrieve=None, pattern=None):
        """Perform recusive file search for all input variables

        Note
        ----
        Overloaded implementation of base class, since for Earlinet, the
        paths are variable dependent

        Parameters
        ----------
        vars_to_retrieve : list
            list of variables to retrieve
        pattern : str, optional
            file name pattern applied to search

        Returns
        -------
        list
            list containing file paths
        """

        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        exclude = self._get_exclude_filelist()
        logger.info("Fetching EARLINET data files. This might take a while...")
        patterns = []
        for var in vars_to_retrieve:
            if var not in self.VAR_PATTERNS_FILE:
                from pyaerocom.exceptions import VarNotAvailableError

                raise VarNotAvailableError(f"Input variable {var} is not supported")

            _pattern = self.VAR_PATTERNS_FILE[var]
            if pattern is not None:
                if "." in pattern:
                    raise NotImplementedError("filetype delimiter . not supported")
                spl = _pattern.split(".")
                if "*" not in spl[0]:
                    raise AttributeError(f"Invalid file pattern: {_pattern}")
                spl[0] = spl[0].replace("*", pattern)
                _pattern = ".".join(spl)

            patterns.append(_pattern)

        matches = []
        for root, dirnames, files in os.walk(self.data_dir, topdown=True):
            paths = [os.path.join(root, f) for f in files]
            for _pattern in patterns:
                for path in paths:
                    file = os.path.basename(path)
                    if _pattern not in file:
                        continue
                    elif file in exclude:
                        self.excluded_files.append(path)
                    else:
                        matches.append(path)
        self.files = files = list(dict.fromkeys(matches))
        return files



import pytest

from pyaerocom.io import ReadAeronetSdaV3, ReadAeronetSunV3
from pyaerocom.ungriddeddata import UngriddedData


@pytest.fixture(scope="session")
def aeronet_sun_subset_reader() -> ReadAeronetSunV3:
    return ReadAeronetSunV3("AeronetSunV3L2Subset.daily")


@pytest.fixture(scope="session")
def aeronet_sda_subset_reader() -> ReadAeronetSdaV3:
    return ReadAeronetSdaV3("AeronetSDAV3L2Subset.daily")


@pytest.fixture(scope="session")
def aeronetsunv3lev2_subset(aeronet_sun_subset_reader: ReadAeronetSunV3) -> UngriddedData:
    return aeronet_sun_subset_reader.read(vars_to_retrieve=["od550aer", "ang4487aer"])


@pytest.fixture(scope="session")
def aeronetsdav3lev2_subset(aeronet_sda_subset_reader: ReadAeronetSdaV3) -> UngriddedData:
    return aeronet_sda_subset_reader.read(vars_to_retrieve=["od550aer", "od550lt1aer"])



##################################################
#        The global configs
##################################################
from pathlib import Path

GLOBAL_CONFIG = dict(
    # Description of the experiment
    proj_id="cams2-83",  # cannot be changed because it has a role in determining the output subfolders
    exp_id="prototype",
    exp_name="Prototype-daily",
    exp_descr=("Evaluation using EEA-MF NRT obs."),
    exp_pi="<a href='https://atmosphere.copernicus.eu/help-and-support'>CAMS user support</a>",
    # Whether or not the experiment is visible in the web interface
    public=True,
    # Locations where to place the results
    # These can be set as the user want, but as here written to use the folder structures we made
    json_basedir=str(Path("../../data").absolute()),
    coldata_basedir=str(Path("../../coldata").absolute()),
    # io_aux_file=os.path.abspath("../eval_py/gridded_io_aux.py"),
    # Some info about the output
    reanalyse_existing=True,
    only_json=False,
    add_model_maps=False,
    # maps_res_deg=5,
    only_model_maps=False,
    clear_existing_json=False,
    # if True, the analysis will stop whenever an error occurs (else, errors that
    # occurred will be written into the logfiles)
    raise_exceptions=False,
    # options for CAMS2-83
    use_cams2_83=True,
    # cams2_83_model=ModelName.EMEP,
    # cams2_83_dateshift=0,
    # Regional filter for analysis
    filter_name="ALL-wMOUNTAINS",
    # colocation frequency (no statistics in higher resolution can be computed)
    ts_type="hourly",
    # The size of map used to display the results
    map_zoom="Europe",
    # Options for time
    freqs=["hourly", "daily"],  # Possible frequencies
    periods=[
        "2021-2022"
    ],  # Periodes, can be single years or range, e.g. 2010-2015. EMEP only supports single years as of now
    main_freq="hourly",  # default frequency to use. This will be overwritten in most of the observation options (see below)
    # This has to be true for the web interface to show diurnal evaluation
    use_diurnal=False,
    # O3 is special, since we want to look at daily max
    # Here we say that we when O3(vmro3) is evaluated, the daily results will be the maximum for that day
    resample_how={"vmro3": {"daily": {"hourly": "max"}}},
    # Assorted options, more info can be found in 'cfg_examples_examples1.py'
    # zeros_to_nan=False,
    zeros_to_nan=True,
    colocate_time=False,
    obs_remove_outliers=False,
    model_remove_outliers=False,
    harmonise_units=True,
    regions_how="country",
    annual_stats_constrained=False,
    weighted_stats=False,
    forecast_evaluation=True,
    forecast_days=4,
    use_fairmode=False,
    drop_stats=("mb", "mab"),
    # This is just the order at which the different species will be shown in the web interface
    # Species that are not evaluated can still be in this list. The web interface will not show them if they are not evaluated
    var_order_menu=[
        "concco",
        "concno2",        
        "conco3",
        "conco3mda8",
        "concpm10",
        "concpm25",
        "concso2",
    ],
    min_num_obs=dict(
        # yearly=dict(monthly=9),
        # monthly=dict(daily=21, weekly=3),
        daily=dict(hourly=18),
    ),
)


##################################################
#        The model configs
##################################################

MODELS_CONFIG = {}


##################################################
#        The observation configs
##################################################
# Station filters
ignore_id_dict = dict(
    concso2="GR0001*",
    vmro3="RS0005*",
    concNnh4="EE0009*",
    concso4="EE0009*",
    concco=["BETN*"],
)

BASE_FILTER = {
    "latitude": [30, 82],
    "longitude": [-30, 90],
}

EEA_RURAL_FILTER = {
    "station_classification": ["background"],
    "area_classification": [
        "rural",
        "rural-nearcity",
        "rural-regional",
        "rural-remote",
    ],
}
EBAS_FILTER = {
    "data_level": [None, 2],
    **BASE_FILTER,
    "altitude": [-20, 1000],
    "set_flags_nan": True,
}

EEA_FILTER = {
    **BASE_FILTER,
    **EEA_RURAL_FILTER,
    "altitude": [-20, 1000],
}

species_list = [
    "concno2",
    "concco",
    "conco3",
    "concso2",
    "concpm10",
    "concpm25",
]


def get_ignore_list(species):
    return ignore_id_dict[species] if species in ignore_id_dict else ["NO0042*"]


obs_filters = {
    key: dict(
        **BASE_FILTER,
        station_id=get_ignore_list(key),
        negate="station_id",
    )
    for key in species_list
}

# Empty observation config
OBS_CONFIG = {}

# EEA observatio
OBS_CONFIG["EEA"] = dict(
    obs_id="CAMS2_83.NRT",
    # obs_id="EEAAQeRep.NRT",
    obs_vars=species_list,
    web_interface_name="EEA-UTD",
    obs_vert_type="Surface",
    read_opts_ungridded=dict(files=[], force_caching=True),
    obs_filters=obs_filters,
)

##################################################
#        Putting it all together
##################################################
CFG = dict(
    model_cfg=MODELS_CONFIG,
    obs_cfg=OBS_CONFIG,
    **GLOBAL_CONFIG,
)



import sys

import numpy as np
import pytest

from pyaerocom.io.ebas_nasa_ames import EbasColDef, EbasFlagCol, EbasNasaAmesFile, NasaAmesHeader


@pytest.fixture(scope="module")
def head():
    return NasaAmesHeader


def test_EbasFlagCol():
    fc = EbasFlagCol(np.asarray([0, 0.66, 0.456, 0.999, 0.999100, 0.999100456]))
    assert (fc.valid == np.asarray([True, True, False, False, True, True])).all()


@pytest.mark.parametrize(
    "raw_data,decoded",
    [
        (np.asarray([0]), np.asarray([[0, 0, 0]])),
        (np.asarray([0.10045666]), np.asarray([[100, 456, 660]])),
        (np.asarray([0.10045666, 0]), np.asarray([[100, 456, 660], [0, 0, 0]])),
        (np.asarray([0.10045666, 0.12]), np.asarray([[100, 456, 660], [120, 0, 0]])),
        (np.asarray([0.10045666, 0.1234]), np.asarray([[100, 456, 660], [123, 400, 0]])),
    ],
)
def test_EbasFlagCol_decoded(raw_data, decoded):
    fc = EbasFlagCol(raw_data, False)
    assert fc._decoded is None
    dc = fc.decoded
    assert fc._decoded is dc
    assert decoded.ndim == 2
    assert (dc == decoded).all()


def test_NasaAmesHeader_NUM_FIXLINES(head):
    assert head._NUM_FIXLINES == 13


def test_NasaAmesHeader_CONV_STR(head):
    assert head.CONV_STR("bla ") == "bla"


def test_NasaAmesHeader_CONV_PI(head):
    assert head.CONV_PI("bla;blub") == "bla; blub"


def test_EbasNasaAmesFile_instance(loaded_nasa_ames_example: EbasNasaAmesFile):
    assert isinstance(loaded_nasa_ames_example, NasaAmesHeader)
    assert isinstance(loaded_nasa_ames_example, EbasNasaAmesFile)


def test_EbasNasaAmesFile_head_fix(loaded_nasa_ames_example: EbasNasaAmesFile):
    HEAD_FIX = dict(
        num_head_lines=93,
        num_head_fmt=1001,
        data_originator="Brem, Benjamin; Baltensperger, Urs",
        sponsor_organisation="CH02L, Paul Scherrer Institut, PSI, Laboratory of Atmospheric Chemistry (LAC), OFLB, , 5232, Villigen PSI, Switzerland",
        submitter="Brem, Benjamin",
        project_association="ACTRIS CREATE EMEP GAW-WDCA",
        vol_num=1,
        vol_totnum=1,
        ref_date=np.datetime64("2019-01-01T00:00:00"),
        revision_date=np.datetime64("2021-05-28T00:00:00"),
        freq=0.041667,
        descr_time_unit="days from file reference point",
        num_cols_dependent=23,
        mul_factors=[1.0] * 23,
        vals_invalid=[999.999999, 9999.99, 999.99, 9999.99]
        + [99.99999999] * 9
        + [999.99999999] * 9
        + [9.999999],
        descr_first_col="end_time of measurement, days from the file reference point",
    )
    assert isinstance(loaded_nasa_ames_example.head_fix, dict)
    assert loaded_nasa_ames_example.head_fix == HEAD_FIX


def test_EbasNasaAmesFile_head_fix_error(loaded_nasa_ames_example: EbasNasaAmesFile):
    with pytest.raises(AttributeError) as e:
        loaded_nasa_ames_example.head_fix = "Blaaaaaaaaaaaaaaa"
    if sys.version_info < (3, 11):
        assert str(e.value).startswith("can't set attribute")
    else:
        assert str(e.value).endswith("object has no setter")


def test_EbasNasaAmesFile_data(loaded_nasa_ames_example: EbasNasaAmesFile):
    assert isinstance(loaded_nasa_ames_example.data, np.ndarray)
    assert loaded_nasa_ames_example.data.ndim == 2


def test_EbasNasaAmesFile_shape(loaded_nasa_ames_example: EbasNasaAmesFile):
    assert loaded_nasa_ames_example.shape == (8760, 24)


def test_EbasNasaAmesFile_col_num(loaded_nasa_ames_example: EbasNasaAmesFile):
    assert loaded_nasa_ames_example.col_num == 24


def test_EbasNasaAmesFile_col_names(loaded_nasa_ames_example: EbasNasaAmesFile):
    COLUMN_NAMES = (
        ["starttime", "endtime", "pressure", "relative_humidity", "temperature"]
        + ["aerosol_light_backscattering_coefficient"] * 9
        + ["aerosol_light_scattering_coefficient"] * 9
        + ["numflag"]
    )
    assert loaded_nasa_ames_example.col_names == COLUMN_NAMES


def test_EbasNasaAmesFile_get_time_gaps_meas(loaded_nasa_ames_example: EbasNasaAmesFile):
    gaps = loaded_nasa_ames_example.get_time_gaps_meas()
    assert len(gaps) == 8759
    assert np.unique(gaps).sum() == 0


def test_EbasNasaAmesFile_get_dt_meas(loaded_nasa_ames_example: EbasNasaAmesFile):
    dt = loaded_nasa_ames_example.get_dt_meas()
    assert len(dt) == 8759
    assert list(np.unique(dt)) == [3599.0, 3600.0, 3601.0]


@pytest.mark.parametrize("update", [{"bla": 42}, {"vol_num": 42}])
def test_EbasNasaAmesFile_update(loaded_nasa_ames_example: EbasNasaAmesFile, update: dict):
    data = EbasNasaAmesFile(loaded_nasa_ames_example.file)
    data.update(**update)
    for key, val in update.items():
        if key in loaded_nasa_ames_example._head_fix:
            assert data._head_fix[key] == val
        else:
            assert data._meta[key] == val


def test_EbasNasaAmesFile___str__(loaded_nasa_ames_example: EbasNasaAmesFile):
    assert isinstance(loaded_nasa_ames_example.__str__(), str)


@pytest.mark.parametrize(
    "colnum,value",
    [
        (5, 450),
        (8, 550),
    ],
)
def test_EbasColDef_get_wavelength_nm(
    loaded_nasa_ames_example: EbasNasaAmesFile, colnum: int, value: int
):
    coldef = loaded_nasa_ames_example.var_defs[colnum]
    assert isinstance(coldef, EbasColDef)
    assert coldef.get_wavelength_nm() == value


def test_EbasColDef_get_wavelength_nm_error(loaded_nasa_ames_example: EbasNasaAmesFile):
    with pytest.raises(KeyError) as e:
        loaded_nasa_ames_example.var_defs[0].get_wavelength_nm()
    assert str(e.value) == "'Column variable starttime does not contain wavelength information'"



"""
Methods to convert different standards of vertical coordinates

For details see here:

http://cfconventions.org/Data/cf-conventions/cf-conventions-1.0/build/apd.html

Note
----
UNDER DEVELOPMENT -> NOT READY YET
"""

import logging

import numpy as np

from pyaerocom import const
from pyaerocom.exceptions import (
    CoordinateNameError,
    VariableDefinitionError,
)

logger = logging.getLogger(__name__)


def atmosphere_sigma_coordinate_to_pressure(
    sigma: np.ndarray | float, ps: float, ptop: float
) -> np.ndarray | float:
    """Convert atmosphere sigma coordinate to pressure in Pa

    Note
    ----
    This formula only works at one lon lat coordinate and at one instant in
    time.

    **Formula**:

    .. math::

        p(k) = p_{top} + \\sigma(k) \\cdot (p_{surface} - p_{top})

    Parameters
    ----------
    sigma : ndarray or float
        sigma coordinate (1D) array
    ps : float
        surface pressure
    ptop : float
        ToA pressure

    Returns
    -------
    ndarray or float
        computed pressure levels in Pa (standard_name=air_pressure)
    """
    if not isinstance(ptop, float | np.ndarray):
        try:
            ptop = float(ptop)
        except ValueError as e:
            raise ValueError(
                f"Invalid input for ptop. Need floating point\nError: {repr(e)}"
            ) from e
    return ptop + sigma * (ps - ptop)


def atmosphere_hybrid_sigma_pressure_coordinate_to_pressure(
    a: np.ndarray, b: np.ndarray, ps: float, p0: float | None = None
) -> np.ndarray:
    """Convert atmosphere_hybrid_sigma_pressure_coordinate to  pressure in Pa

    **Formula**:

    Either

    .. math::

        p(k) = a(k) \\cdot p_0 + b(k) \\cdot p_{surface}

    or

    .. math::

        p(k) = ap(k) + b(k) \\cdot p_{surface}

    Parameters
    ----------
    a : ndarray
        sigma level values (a(k) in formula 1, and ap(k) in formula 2)
    b : ndarray
        dimensionless fraction per level (must be same length as a)
    ps : float
        surface pressure
    p0 :
        reference pressure (only relevant for alternative formula 1)

    Returns
    -------
    ndarray
        computed pressure levels in Pa (standard_name=air_pressure)

    """
    if len(a) != len(b):
        raise ValueError("Invalid input: a and b must have the same length")
    if p0 is None:
        return a + b * ps
    return a * p0 + b * ps


class VerticalCoordinate:
    NAMES_SUPPORTED = {
        "altitude": "z",
        "air_pressure": "pres",
        "geopotential_height": "gph",
        "atmosphere_sigma_coordinate": "asc",
        "atmosphere_hybrid_sigma_pressure_coordinate": "ahspc",
    }

    STANDARD_NAMES = dict(map(reversed, NAMES_SUPPORTED.items()))

    NAMES_NOT_SUPPORTED = ["model_level_number"]

    #: registered names
    REGISTERED = list(NAMES_SUPPORTED) + NAMES_NOT_SUPPORTED

    CONVERSION_METHODS = {
        "asc": atmosphere_sigma_coordinate_to_pressure,
        "ahspc": atmosphere_hybrid_sigma_pressure_coordinate_to_pressure,
    }

    CONVERSION_REQUIRES = {
        "asc": ["sigma", "ps", "ptop"],
        "ahspc": ["a", "b", "ps", "p0"],
        "gph": [],
    }

    FUNS_YIELD = {"asc": "air_pressure", "ahspc": "air_pressure", "gph": "altitude"}

    _LEV_INCREASES_WITH_ALT = dict(
        atmosphere_sigma_coordinate=False,
        atmosphere_hybrid_sigma_pressure_coordinate=False,
        altitude=True,
    )

    def __init__(self, name=None):
        self.var_name = None
        self.standard_name = None
        self._eval_input(name)

    def _eval_input(self, name):
        if name in self.NAMES_SUPPORTED:
            self.var_name = self.NAMES_SUPPORTED[name]
            self.standard_name = name
        elif name in self.STANDARD_NAMES:
            self.var_name = name
            self.standard_name = self.STANDARD_NAMES[name]
        elif name in self.NAMES_NOT_SUPPORTED:
            self.standard_name = name
        else:
            raise ValueError(f"Invalid name for vertical coordinate: {name}")

    @property
    def fun(self):
        """Function used to convert levels into pressure"""
        return self.CONVERSION_METHODS[self.var_name]

    @property
    def conversion_requires(self):
        """Valid argument names for :func:`fun`"""
        return self.CONVERSION_REQUIRES[self.var_name]

    @property
    def conversion_supported(self) -> bool:
        """Boolean specifying whether a conversion scheme is provided"""
        return True if self.standard_name in self.NAMES_SUPPORTED else False

    @property
    def lev_increases_with_alt(self) -> bool:
        """Boolean specifying whether coordinate levels increase with altitude

        Returns
        -------
        True
        """
        if self.var_name not in self._LEV_INCREASES_WITH_ALT:
            if self.standard_name in self._LEV_INCREASES_WITH_ALT:
                return self._LEV_INCREASES_WITH_ALT[self.standard_name]
            raise ValueError(
                f"Failed to access information lev_increases_with_alt "
                f"for vertical coordinate {self.var_name}"
            )
        return self._LEV_INCREASES_WITH_ALT[self.var_name]

    def calc_pressure(self, lev: np.ndarray, **kwargs) -> np.ndarray:
        """Compute pressure levels for input vertical coordinate

        Parameters
        ----------
        vals : ndarray
            level values that are supposed to be converted into pressure
        **kwargs
            additional  keyword args required for computation of pressure
            levels (cf. :attr:`CONVERSION_METHODS` and corresponding inputs for method
            available)

        Returns
        -------
        ndarray
            pressure levels in Pa
        """

        if self.var_name not in self.NAMES_SUPPORTED:
            raise CoordinateNameError(
                f"Variable {self.var_name} cannot be converted to pressure levels. "
                f"Conversion is only possible for supported variables:\n{self.vars_supported_str}"
            )

        coord_values = kwargs.pop(self.var_name)
        return self.fun(coord_values, **kwargs)

    @property
    def vars_supported_str(self) -> str:
        from pyaerocom._lowlevel_helpers import dict_to_str

        return dict_to_str(self.NAMES_SUPPORTED)

    def pressure2altitude(self, p, **kwargs):
        """Convert pressure to altitude

        Wrapper for method

        Parameters
        ----------

        Returns
        -------
        """
        raise NotImplementedError("Coming soon...")


class AltitudeAccess:
    #: Additional variable names (in AEROCOM convention) that are used
    #: to search for additional files that can be used to access or compute
    #: the altitude levels at each grid point
    ADD_FILE_VARS = ["z", "z3d", "pres", "deltaz3d"]

    #: Additional variables that are required to compute altitude levels
    ADD_FILE_REQ = {"deltaz3d": ["ps"]}

    ADD_FILE_OPT = {"pres": ["temp"]}

    def __init__(self, gridded_data):
        from pyaerocom.griddeddata import GriddedData

        if not isinstance(gridded_data, GriddedData):
            raise ValueError("Invalid input: require instance of GriddedData class")
        if not gridded_data.has_latlon_dims:
            raise NotImplementedError(
                f"Altitude access requires latitude and longitude dimensions to be available "
                f"in input GriddedData: {gridded_data.short_str()}"
            )
        self.data_obj = gridded_data
        self._subset1d = None
        self._checked_and_failed = []
        self.logger = logger

    def __setitem__(self, key, val):
        self.__dict__[key] = val

    def __getitem__(self, key):
        return self.__dict__[key]

    def __contains__(self, key):
        return True if key in self.__dict__ else False

    @property
    def coord_list(self):
        """List of AeroCom coordinate names for altitude access"""
        name = self.ADD_FILE_VARS + list(VerticalCoordinate.NAMES_SUPPORTED.values())
        return sorted(set(name))

    def search_aux_coords(self, coord_list) -> bool:
        """Search and assign coordinates provided by input list

        All coordinates that are found are assigned to this object and can
        be accessed via ``self[coord_name]``.

        Parameters
        ----------
        coord_list : list
            list containing AeroCom coordinate names

        Returns
        -------
        bool
            True if all coordinates can be accessed, else False

        Raises
        ------
        CoordinateNameError
            if one of the input coordinate names is not supported by pyaerocom.
            See coords.ini file of pyaerocom for available coordinates.
        """
        all_ok = True
        if isinstance(coord_list, str):
            coord_list = [coord_list]
        d = self.data_obj
        for coord in coord_list:
            if coord in self:
                continue
            try:
                const.COORDINFO[coord]
            except VariableDefinitionError:
                raise CoordinateNameError(f"Coordinate {coord} is not supported by pyaerocom.")
            # 1. check if coordinate is assigned in data object directly
            d._update_coord_info()
            if coord in d._coord_var_names:
                self[coord] = d[coord]
            else:
                try:
                    self[coord] = d.search_other(coord)
                    print(f"Adding coord {coord}")
                except Exception:
                    all_ok = False
        return all_ok

    @property
    def reader(self):
        """Instance of :class:`ReadGridded`"""
        return self.data_obj.reader



from __future__ import annotations

import logging
from concurrent.futures import ProcessPoolExecutor, as_completed
from datetime import date, timedelta
from enum import Enum
from pathlib import Path
from pprint import pformat

from pyaerocom import const
from pyaerocom.aeroval import EvalSetup, ExperimentProcessor
from pyaerocom.io import ReadUngridded
from pyaerocom.io.cachehandler_ungridded import list_cache_files
from pyaerocom.scripts.cams2_83.config import species_list
from pyaerocom.scripts.cams2_83.processer import CAMS2_83_Processer

logger = logging.getLogger(__name__)


class EvalType(str, Enum):
    LONG = "long"
    SEASON = "season"
    WEEK = "week"
    DAY = "day"

    def __str__(self) -> str:
        return self.value

    def check_dates(self, start_date: date, end_date: date) -> None:
        if self == "day" and start_date != end_date:
            raise ValueError(
                f"Evaluation type 'day' should have the same {start_date=} and {end_date=}"
            )

        if self == "week" and (days := (end_date - start_date) // timedelta(days=1)) < 7:
            raise ValueError(f"Evaluation type 'week' should have {days=} >= 7")

    def freqs_config(self) -> dict:
        if self == "long":
            return dict(
                freqs=["daily", "monthly"],
                ts_type="hourly",
                main_freq="daily",
                forecast_evaluation=True,
            )
        
        if self == "season":
            return dict(
                freqs=["hourly", "daily"],
                ts_type="hourly",
                main_freq="hourly",
                forecast_evaluation=True,
            )

        if self == "week":
            return dict(
                freqs=["hourly", "daily"],
                ts_type="hourly",
                main_freq="hourly",
                forecast_evaluation=False,
            )
        if self == "day":
            return dict(
                freqs=["hourly"],
                ts_type="hourly",
                main_freq="hourly",
                forecast_evaluation=False,
            )

        raise NotImplementedError(f"Unsupported {self}")

    def periods(self, start_date: date, end_date: date) -> list[str]:
        if self == "long":
            if (start_date.year != end_date.year):
                return make_period_ys(start_date, end_date)
        return make_period(start_date, end_date)


def date_range(start_date: date, end_date: date) -> tuple[date, ...]:
    days = (end_date - start_date) // timedelta(days=1)
    assert days >= 0
    return tuple(start_date + timedelta(days=day) for day in range(days + 1))


def make_period(start_date: date, end_date: date) -> list[str]:
    if start_date == end_date:
        return [f"{start_date:%Y%m%d}"]
    periods = [f"{start_date:%Y%m%d}-{end_date:%Y%m%d}"]

    return periods


def make_period_ys(start_date: date, end_date: date) -> list[str]:
    periods = [f"{start_date.year}-{end_date.year}"]
    periods.extend(str(yr) for yr in range(start_date.year, end_date.year + 1))
    return periods


def clear_cache():
    """Delete cached data objects"""
    for path in list_cache_files():
        path.unlink()


def read_observations(specie: str, *, files: list, cache: str | Path | None) -> None:
    logger.info(f"Running {specie}")

    if cache is not None:
        const.CACHEDIR = str(cache)

    reader = ReadUngridded()

    reader.read(
        data_ids="CAMS2_83.NRT",
        vars_to_retrieve=specie,
        files=files,
        force_caching=True,
    )

    logger.info(f"Finished {specie}")


def run_forecast(specie: str, *, stp: EvalSetup, analysis: bool) -> None:
    ana_cams2_83 = CAMS2_83_Processer(stp)
    ana_cams2_83.run(analysis=analysis, var_list=specie)


def runner(
    cfg: dict,
    cache: str | Path | None,
    dry_run: bool = False,
    pool: int = 1,
):
    logger.info(f"Running the evaluation for the config\n{pformat(cfg)}")
    if dry_run:
        return

    if cache is not None:
        const.CACHEDIR = str(cache)

    stp = EvalSetup(**cfg)

    logger.info(f"Clearing cache at {const.CACHEDIR}")
    clear_cache()

    if pool > 1:
        logger.info(f"Running observation reading with pool {pool}")
        files = cfg["obs_cfg"]["EEA"]["read_opts_ungridded"]["files"]
        with ProcessPoolExecutor(max_workers=pool) as executor:
            futures = [
                executor.submit(read_observations, specie, files=files, cache=cache)
                for specie in species_list
            ]
        for future in as_completed(futures):
            future.result()

    logger.info("Running Statistics")
    ExperimentProcessor(stp).run()
    logger.info("Done Running Statistics")


def runnermos(
    cfg: dict,
    cache: str | Path | None,
    dry_run: bool = False,
):
    logger.info(f"Running the evaluation for the config\n{pformat(cfg)}")
    if dry_run:
        return

    if cache is not None:
        const.CACHEDIR = str(cache)

    stp = EvalSetup(**cfg)

    logger.info(f"Clearing cache at {const.CACHEDIR}")
    clear_cache()

    logger.info("Running Statistics (MOS)")
    ExperimentProcessor(stp).run()
    logger.info("Done Running Statistics (MOS)")


def runnermedianscores(
    cfg: dict,
    cache: str | Path | None,
    *,
    analysis: bool = False,
    dry_run: bool = False,
    pool: int = 1,
):
    if dry_run:
        return

    if cache is not None:
        const.CACHEDIR = str(cache)

    stp = EvalSetup(**cfg)

    logger.info(
        "Running CAMS2_83 Specific Statistics, cache is not cleared, colocated data is assumed in place, regular statistics are assumed to have been run"
    )
    if pool > 1:
        logger.info(f"Making median scores plot with pool {pool} and analysis {analysis}")
        with ProcessPoolExecutor(max_workers=pool) as executor:
            futures = [
                executor.submit(run_forecast, specie, stp=stp, analysis=analysis)
                for specie in species_list
            ]
        for future in as_completed(futures):
            future.result()
    else:
        logger.info(f"Making median scores plot with pool {pool} and analysis {analysis}")
        CAMS2_83_Processer(stp).run(analysis=analysis)

    logger.info("Median scores run finished")



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import Normalize
from scipy.stats import kendalltau
from scipy.stats.mstats import theilslopes

from pyaerocom.trends_helpers import (
    _compute_trend_error,
    _get_yearly,
    _init_period_dates,
    _init_trends_result_dict,
    _start_season,
    _start_stop_period,
)


class TrendsEngine:
    """Trend computation engine (does not need to be instantiated)"""

    CMAP = matplotlib.colormaps["bwr"]
    NORM = Normalize(-10, 10)

    @staticmethod
    def compute_trend(
        data, ts_type, start_year, stop_year, min_num_yrs, season=None, slope_confidence=None
    ):
        """
        Compute trend

        Parameters
        ----------
        data : pd.Series
            input timeseries data
        ts_type : str
            frequency of input data (must be monthly or yearly)
        start_year : int or str
            start of period for trend
        stop_year : int or str
            end of period for trend
        min_num_yrs : int
            minimum number of years for trend computation
        season : str, optional
            which season to use, defaults to whole year (no season)
        slope_confidence : float, optional
            confidence of slope, between 0 and 1, defaults to 0.68.

        Returns
        -------
        dict
            trends results for input data

        """

        if season is None:
            season = "all"
        if slope_confidence is None:
            slope_confidence = 0.68
        if ts_type not in ["yearly", "monthly"]:
            raise ValueError(ts_type)

        result = _init_trends_result_dict(start_year)
        start_str = _start_season(season, start_year)
        stop_str = str(stop_year)
        data = data.loc[start_str:stop_str]

        result["period"] = f"{start_year}-{stop_year}"
        result["season"] = season
        if len(data) == 0:
            return result

        (start_date, stop_date, period_index, num_dates_period) = _init_period_dates(
            start_year, stop_year, season
        )

        if ts_type == "monthly":
            data = _get_yearly(data, season, start_year)

        result["data"] = data
        dates = data.index.values
        values = data.values

        # get period filter mask
        tmask = np.logical_and(dates >= start_date, dates <= stop_date)

        # apply period mask to jsdate vector and value vector
        dates_data = dates[tmask]

        # vector containing data values
        vals = values[tmask]

        valid = ~np.isnan(vals)

        # works only on not nan values
        dates_data = dates_data[valid]
        vals = vals[valid]

        num_dates_data = dates_data.astype("datetime64[Y]").astype(np.float64)

        # TODO: len(y) is number of years - 1 due to midseason averages
        result["n"] = len(vals)

        if not len(vals) >= min_num_yrs:
            return result

        result["y_mean"] = np.nanmean(vals)
        result["y_min"] = np.nanmin(vals)
        result["y_max"] = np.nanmax(vals)

        # Mann / Kendall test
        [tau, pval] = kendalltau(x=num_dates_data, y=vals)

        (slope, yoffs, slope_low, slope_up) = theilslopes(
            y=vals, x=num_dates_data, alpha=slope_confidence
        )

        # estimate error of slope at input confidence level
        slope_err = np.mean([abs(slope - slope_low), abs(slope - slope_up)])

        reg_data = slope * num_dates_data + yoffs
        reg_period = slope * num_dates_period + yoffs

        # value used for normalisation of slope to compute trend T
        # T=m / v0
        v0_data = reg_data[0]
        v0_period = reg_period[0]

        # Compute the mean residual value, which is used to estimate
        # the uncertainty in the normalisation value used to compute
        # trend
        mean_residual = np.mean(np.abs(vals - reg_data))

        # trend is slope normalised by first reference value.
        # 2 trends are computed, 1. the trend using the first value of
        # the regression line at the first available data year, 2. the
        # trend corresponding to the value corresponding to the first
        # year of the considered period.

        trend_data = slope / v0_data * 100
        trend_period = slope / v0_period * 100

        # Compute errors of normalisation values
        v0_err_data = mean_residual
        t0_data, tN_data = num_dates_data[0], num_dates_data[-1]
        t0_period = num_dates_period[0]

        # sanity check
        assert t0_data < tN_data
        assert t0_period <= t0_data

        dt_ratio = (t0_data - t0_period) / (tN_data - t0_data)

        v0_err_period = v0_err_data * (1 + dt_ratio)

        trend_data_err = _compute_trend_error(
            m=slope, m_err=slope_err, v0=v0_data, v0_err=v0_err_data
        )

        trend_period_err = _compute_trend_error(
            m=slope, m_err=slope_err, v0=v0_period, v0_err=v0_err_period
        )

        result["pval"] = pval
        result["m"] = slope
        result["m_err"] = slope_err
        result["yoffs"] = yoffs

        result["slp"] = trend_data
        result["slp_err"] = trend_data_err
        result["reg0"] = v0_data
        tp, tperr, v0p = None, None, None
        if v0_period > 0:
            tp = trend_period
            tperr = trend_period_err
            v0p = v0_period
        result[f"slp_{start_year}"] = tp
        result[f"slp_{start_year}_err"] = tperr
        result[f"reg0_{start_year}"] = v0p

        return result


class TrendPlotter:  # pragma: no cover
    def __init__(self):
        raise NotImplementedError(
            "consider removing or re-implementing (based on code from v0.10.0)"
        )

    def get_trend_color(self, trend_val):
        return self.CMAP(self.NORM(trend_val))

    def plot(self, season="all", period=None, ax=None):
        if season not in self.seasons_avail:
            raise AttributeError(f"No results available for season {season}")
        if period is None:
            if len(self.results[season]) > 1:
                raise ValueError(
                    f"Found multiple trends for different periods: {list(self.results[season])}. "
                    f"Please specify period..."
                )
            period = list(self.results[season])[0]
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(18, 8))
        if self.has_daily:
            ax.plot(self.daily, "-", marker=".", label="daily", c="#d9d9d9")
        if self.has_monthly:
            ax.plot(self.monthly, label="monthly", c="#4d4d4d")
        ax.plot(self.get_yearly(season), " ok", label="yearly")
        if period in self.periods_avail:
            (s_data, s_period, td, tp, tdstr, tpstr) = self._get_trend_data(season, period)

            ax.plot(s_data, "-", color=self.get_trend_color(td), label="trend", lw=2)
            ax.plot(s_period, "--", color=self.get_trend_color(tp))
            ax.text(0.01, 0.95, tpstr, transform=ax.transAxes, fontsize=14)
            ax.text(0.01, 0.9, tdstr, transform=ax.transAxes, fontsize=14)

        ax.yaxis.grid(c="#d9d9d9", ls="-.")
        ylbl = self.var_name
        if self.var_name is not None and "units" in self.meta:
            u = str(self.meta["units"])
            if u not in ["", "1"]:
                ylbl += f" [{u}]"
        ax.set_ylabel(ylbl)
        tit = ""
        if self.meta["station_name"] is not None:
            tit += self.meta["station_name"]
            try:
                self.meta.load_dataset_info()
            except Exception:
                pass
            dsinfo = self.meta.dataset_str()
            if dsinfo is not None:
                tit += f"; {dsinfo}"
            tit += " - "
        tit += period
        ax.set_title(tit)
        ax.legend(loc=1)

        ax.set_xlim(_start_stop_period(period))

        return ax



import abc
from collections.abc import Iterator

from pyaerocom.griddeddata import GriddedData


class GriddedReader(abc.ABC):
    """Abstract base class for griddel model reader used for collocation"""

    @property
    @abc.abstractmethod
    def data_id(self) -> str:
        """
        Data ID of dataset
        """
        pass

    @property
    @abc.abstractmethod
    def ts_types(self) -> Iterator[str]:
        """
        List of available frequencies

        Raises
        ------
        AttributeError
            if :attr:`data_dir` is not set.

        Returns
        -------
        list
            list of available frequencies

        """
        pass

    @property
    @abc.abstractmethod
    def years_avail(self) -> Iterator[str]:
        """
        Years available in dataset
        """
        pass

    @property
    @abc.abstractmethod
    def vars_provided(self) -> Iterator[str]:
        """Variables provided by this dataset"""
        pass

    @abc.abstractmethod
    def has_var(self, var_name) -> bool:
        """Check if variable is supported

        Parameters
        ----------
        var_name : str
            variable to be checked

        Returns
        -------
        bool
        """
        pass

    @abc.abstractmethod
    def read_var(self, var_name, ts_type=None, **kwargs) -> GriddedData:
        """Load data for given variable.

        Parameters
        ----------
        var_name : str
            Variable to be read
        ts_type : str
            Temporal resolution of data to read. Supported are
            "hourly", "daily", "monthly" , "yearly".

        Returns
        -------
        GriddedData
        """
        pass



from __future__ import annotations

from pathlib import Path

import pytest

from pyaerocom import const
from pyaerocom.io.cnemc.reader import ReadCNEMC

STATION_NAMES = ("1478A", "2706A", "3377A")
VARS_DEFAULT = {"concco", "concno2", "conco3", "concpm10", "concpm25", "concso2"}
VARS_PROVIDED = VARS_DEFAULT | {"vmro3", "vmro3max", "vmrno2"}


@pytest.fixture(scope="module")
def mep_path() -> Path:
    try:
        path = Path(const.OBSLOCS_UNGRIDDED[const.CNEMC_NAME])
    except KeyError:
        pytest.skip("CNEMC path is not registered")

    if not path.exists():
        pytest.skip(f"missing {path}, this os OK on CI")

    return path


@pytest.fixture(scope="module")
def reader(mep_path: Path) -> ReadCNEMC:
    return ReadCNEMC(data_dir=str(mep_path))


@pytest.fixture()
def station_files(mep_path: Path, station: str) -> list[Path]:
    files = sorted(mep_path.rglob(f"mep-rd-{station}-*.nc"))
    assert files, f"no files for {station}"
    return files


def test_DATASET_NAME(reader: ReadCNEMC):
    assert reader.DATASET_NAME == "CNEMC"


def test_DEFAULT_VARS(reader: ReadCNEMC):
    assert set(reader.DEFAULT_VARS) >= VARS_DEFAULT


def test_files(reader: ReadCNEMC):
    assert reader.files, "no stations files found"
    assert len(reader.files) >= 3, "found less files than expected"


def test_FOUND_FILES(reader: ReadCNEMC):
    assert reader.FOUND_FILES, "no stations files found"
    assert len(reader.FOUND_FILES) >= 3, "found less files than expected"


@pytest.mark.parametrize("station", STATION_NAMES)
def test_stations(reader: ReadCNEMC, station: str):
    assert reader.stations()[station], f"no {station} station files"


def test_PROVIDES_VARIABLES(reader: ReadCNEMC):
    assert set(reader.PROVIDES_VARIABLES) >= VARS_PROVIDED


@pytest.mark.parametrize("station", STATION_NAMES)
def test_read_file(reader: ReadCNEMC, station_files: list[str]):
    data = reader.read_file(station_files[-1])
    assert set(data.contains_vars) == VARS_DEFAULT


def test_read_file_error(reader: ReadCNEMC):
    bad_station_file = "not-a-file.nc"
    with pytest.raises(ValueError) as e:
        reader.read_file(bad_station_file)
    assert str(e.value) == f"missing {bad_station_file}"


@pytest.mark.parametrize("station", STATION_NAMES)
def test_read(reader: ReadCNEMC, station_files: list[str]):
    data = reader.read(VARS_PROVIDED, station_files, first_file=0, last_file=5)
    assert set(data.contains_vars) == VARS_PROVIDED


def test_read_error(reader: ReadCNEMC):
    bad_variable_name = "not-a-variable"
    with pytest.raises(ValueError) as e:
        reader.read((bad_variable_name,))
    assert str(e.value) == f"Unsupported variables: {bad_variable_name}"


def test_reader_gives_correct_mep_path(reader: ReadCNEMC, mep_path: Path):
    assert Path(reader.data_dir) == mep_path



import logging
import os
import warnings
from copy import deepcopy
from pathlib import Path

from pyaerocom import const
from pyaerocom.combine_vardata_ungridded import combine_vardata_ungridded
from pyaerocom.exceptions import DataRetrievalError, NetworkNotImplemented, NetworkNotSupported
from pyaerocom.helpers import varlist_aerocom
from pyaerocom.io import ReadUngriddedBase
from pyaerocom.io.cachehandler_ungridded import CacheHandlerUngridded
from pyaerocom.io.cams2_83.read_obs import ReadCAMS2_83
from pyaerocom.io.cnemc.reader import ReadCNEMC
from pyaerocom.io.gaw.reader import ReadGAW
from pyaerocom.io.ghost.reader import ReadGhost
from pyaerocom.io.icos.reader import ReadICOS
from pyaerocom.io.icpforests.reader import ReadICPForest
from pyaerocom.io.pyaro.pyaro_config import PyaroConfig
from pyaerocom.io.pyaro.read_pyaro import ReadPyaro
from pyaerocom.io.read_aasetal import ReadAasEtal
from pyaerocom.io.read_aeronet_invv3 import ReadAeronetInvV3
from pyaerocom.io.read_aeronet_sdav3 import ReadAeronetSdaV3
from pyaerocom.io.read_aeronet_sunv3 import ReadAeronetSunV3
from pyaerocom.io.read_airnow import ReadAirNow
from pyaerocom.io.read_earlinet import ReadEarlinet
from pyaerocom.io.read_ebas import ReadEbas
from pyaerocom.io.read_eea_aqerep import ReadEEAAQEREP
from pyaerocom.io.read_eea_aqerep_v2 import ReadEEAAQEREP_V2
from pyaerocom.ungriddeddata import UngriddedData
from pyaerocom.variable import get_aliases

logger = logging.getLogger(__name__)


# TODO Add check if data id of config is same as one already given by pyaerocom
class ReadUngridded:
    """Factory class for reading of ungridded data based on obsnetwork ID

    This class also features reading functionality that goes beyond reading
    of inidividual observation datasets; including, reading of multiple
    datasets and post computation of new variables based on datasets that can
    be read.

    Parameters
    ----------
    COMING SOON

    """

    SUPPORTED_READERS = [
        ReadAeronetInvV3,
        ReadAeronetSdaV3,
        ReadAeronetSunV3,
        ReadEarlinet,
        ReadEbas,
        ReadAasEtal,
        ReadAirNow,
        ReadEEAAQEREP,
        ReadEEAAQEREP_V2,
        ReadCAMS2_83,
        ReadGAW,
        ReadGhost,
        ReadCNEMC,
        ReadICOS,
        ReadICPForest,
    ]

    # Creates list of all readers excluding ReadPyaro
    INCLUDED_READERS = deepcopy(SUPPORTED_READERS)

    # Adds ReadPyaro to said list
    SUPPORTED_READERS.append(ReadPyaro)

    DONOTCACHE_NAME = "DONOTCACHE"

    def __init__(
        self,
        data_ids=None,
        ignore_cache=False,
        data_dirs=None,
        configs: PyaroConfig | list[PyaroConfig] | None = None,
    ):
        # will be assigned in setter method of data_ids
        self._data_ids = []
        self._data_dirs = {}

        #: dictionary containing reading classes for each dataset to read (will
        #: be accessed via get_reader)
        self._readers = {}

        if data_ids is not None:
            self.data_ids = data_ids

        if data_dirs is not None:
            self.data_dirs = data_dirs

        if ignore_cache:
            logger.info("Deactivating caching")
            const.CACHING = False

        self.config_ids = {}
        self.config_map = {}

        if isinstance(configs, PyaroConfig):
            configs = [configs]

        self._configs = configs

        if isinstance(configs, list):
            for config in configs:
                if config is not None:
                    self._init_pyaro_reader(config=config)

    @property
    def data_dirs(self):
        """
        dict: Data directory(ies) for dataset(s) to read (keys are data IDs)
        """
        return self._data_dirs

    @data_dirs.setter
    def data_dirs(self, val):
        if isinstance(val, Path):
            val = str(val)
        dsr = self.data_ids
        if len(dsr) < 2 and isinstance(val, str):
            val = {dsr[0]: val}
        elif not isinstance(val, dict):
            raise ValueError(f"Invalid input for data_dirs ({val}); needs to be a dictionary.")
        for data_dir in val.values():
            assert os.path.exists(data_dir), f"{data_dir} does not exist"
        self._data_dirs = val

    @property
    def post_compute(self):
        """Information about datasets that can be computed in post"""
        return const.OBS_UNGRIDDED_POST

    @property
    def INCLUDED_DATASETS(self):
        lst = []
        for reader in self.INCLUDED_READERS:
            lst.extend(reader.SUPPORTED_DATASETS)
        lst.extend(self.post_compute)
        return lst

    @property
    def SUPPORTED_DATASETS(self):
        """
        Returns list of strings containing all supported dataset names
        """
        lst = []
        for reader in self.SUPPORTED_READERS:
            lst.extend(reader.SUPPORTED_DATASETS)
        lst.extend(self.post_compute)
        lst.extend(list(self.config_ids.keys()))
        return lst

    @property
    def supported_datasets(self):
        """
        Wrapper for :attr:`SUPPORTED_DATASETS`
        """
        return self.SUPPORTED_DATASETS

    def _check_donotcachefile(self):
        """Check if donotcache file exists

        Returns
        -------
        bool
            True if file exists, else False
        """
        try:
            if os.path.exists(os.path.join(const.cache_basedir, self.DONOTCACHE_NAME)):
                return True
        except Exception:
            pass
        return False

    @property
    def ignore_cache(self):
        """Boolean specifying whether caching is active or not"""
        if self._check_donotcachefile() or not const.CACHING:
            return True
        return False

    @property
    def data_ids(self):
        """List of datasets supposed to be read"""
        return self._data_ids

    @data_ids.setter
    def data_ids(self, val):
        if isinstance(val, str):
            val = [val]
        elif not isinstance(val, tuple | list):
            raise OSError("Invalid input for parameter data_ids")
        self._data_ids = val

    @property
    def configs(self):
        """List configs"""
        return self._configs

    @configs.setter
    def configs(self, val: PyaroConfig | list[PyaroConfig]):
        if isinstance(val, PyaroConfig):
            val = [val]
        elif not isinstance(val, tuple | list):
            raise OSError("Invalid input for parameter data_ids")
        logger.warning(
            "You are now overwriting the list of configs. This will delete the previous configs, but will leave readeres associated with those configs intact. Use 'add_config' for safer usage!"
        )
        for config in val:
            self._init_pyaro_reader(config=config)
        self._configs = deepcopy(val)

    @property
    def data_id(self):
        """
        ID of dataset

        Note
        -----
        Only works if exactly one dataset is assigned to the reader, that is,
        length of :attr:`data_ids` is 1.

        Raises
        ------
        AttributeError
            if number of items in :attr:`data_ids` is unequal one.

        Returns
        -------
        str
            data ID

        """
        nids = len(self.data_ids)
        if nids == 0:
            raise AttributeError("No data_id assigned")
        elif nids > 1:
            raise AttributeError("Multiple data_ids assigned")
        return self.data_ids[0]

    def dataset_provides_variables(self, data_id=None):
        """List of variables provided by a certain dataset"""
        if data_id is None:
            data_id = self.data_id
        if data_id not in self._readers:
            reader = self.get_lowlevel_reader(data_id)
        else:
            reader = self._readers[data_id]
        return reader.PROVIDES_VARIABLES

    def get_reader(self, data_id):
        warnings.warn(
            "this method was renamed to get_lowlevel_reader, please use the new name",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.get_lowlevel_reader(data_id)

    def get_lowlevel_reader(self, data_id: str | None = None) -> ReadUngriddedBase:
        """Helper method that returns initiated reader class for input ID

        Parameters
        -----------
        data_id : str
            Name of dataset

        Returns
        -------
        ReadUngriddedBase
            instance of reading class (needs to be implementation of base
            class :class:`ReadUngriddedBase`).
        """

        if data_id is None:
            if len(self.data_ids) != 1:
                raise ValueError("Please specify dataset")
        if data_id not in self.supported_datasets:
            if data_id not in self.config_map:
                raise NetworkNotSupported(
                    f"Could not fetch reader class: Input "
                    f"network {data_id} is not supported by "
                    f"ReadUngridded"
                )
        elif data_id not in self.data_ids:
            self.data_ids.append(data_id)
        if data_id not in self._readers:
            _cls = self._find_read_class(data_id)
            reader = self._init_lowlevel_reader(_cls, data_id)
            self._readers[data_id] = reader
        return self._readers[data_id]

    def add_pyaro_reader(self, config: PyaroConfig) -> ReadUngriddedBase:
        return self._init_pyaro_reader(config=config)

    def _init_pyaro_reader(self, config: PyaroConfig) -> ReadUngriddedBase:
        """
        Initializes PyAro reader from config, and adds reader to list of readers. If no config is given, the config given when ReaderUngridded was initiated is used

        Parameters
        -----------
        config : PyaroConfig
            Config for reader

        Returns
        -------
        ReadUngriddedBase
            instance of reading class (needs to be implementation of base
            class :class:`ReadUngriddedBase`)


        Raises
        ------
        ValueError
            If both the config argument and self.config are None
        """
        name = config.name

        if name in self.INCLUDED_DATASETS:
            raise NameError(
                f"{name} from config {config} cannot have the same name as an included dataset"
            )

        if name in self._readers:
            return self._readers[name]

        else:
            reader = ReadPyaro(config=config)
            self._readers[name] = reader
            self._data_ids.append(name)
            self.config_ids[name] = config.reader_id
            self.config_map[name] = config
            return reader

    def add_config(self, config: PyaroConfig) -> None:
        """
        Adds single PyaroConfig to self.configs

        Parameters
        ----------
        config: PyaroConfig

        Raises
        ------
        ValueError
            If config is not PyaroConfig

        """
        if not isinstance(config, PyaroConfig):
            raise ValueError("Given config is not a PyaroConfig")

        self._init_pyaro_reader(config=config)
        self._configs.append(config)

    def _find_read_class(self, data_id):
        """Find reading class for dataset name

        Loops over all reading classes available in :attr:`SUPPORTED_READERS`
        and finds the first one that matches the input dataset name, by
        checking the attribute :attr:`SUPPORTED_DATASETS` in each respective
        reading class.

        Parameters
        -----------
        data_id : str
            Name of dataset

        Returns
        -------
        ReadUngriddedBase
            instance of reading class (needs to be implementation of base
            class :class:`ReadUngriddedBase`)

        Raises
        ------
        NetworkNotImplemented
            if network is supported but no reading routine is implemented yet

        """
        for _cls in self.SUPPORTED_READERS:
            if data_id in _cls.SUPPORTED_DATASETS:
                return _cls
        raise NetworkNotImplemented(f"Could not find reading class for dataset {data_id}")

    def _init_lowlevel_reader(self, reader, data_id):
        """
        Initiate lowlevel reader for input data ID

        Parameters
        ----------
        reader
            reader class (not instantiated)
        data_id : str
            ID of dataset to be isntantiated with input reader

        Returns
        -------
        ReadUngriddedBase
            instantiated reader class for input ID.

        """
        # if data_id is not None and config is not None:
        #     if data_id != config.name:
        #         raise ValueError(
        #             f"DATA ID and config are both given, but they are not equal, {data_id} != {config.data_id}"
        #         )
        # if config is None:
        #     config = self.config

        # if data_id is None:
        #     data_id = config.name

        if data_id is None:
            raise ValueError("Data_id can not be none")

        if data_id in self.config_map:
            return reader(config=self.config_map[data_id])

        if data_id in self.data_dirs:
            ddir = self.data_dirs[data_id]
            logger.info(f"Reading {data_id} from specified data loaction: {ddir}")
        else:
            ddir = None

        return reader(data_id=data_id, data_dir=ddir)

    def read_dataset(
        self,
        data_id,
        vars_to_retrieve=None,
        only_cached=False,
        filter_post=None,
        **kwargs,
    ):
        """Read dataset into an instance of :class:`ReadUngridded`

        Parameters
        ----------
        data_id : str
            name of dataset
        vars_to_retrieve : list
            variable or list of variables to be imported
        only_cached : bool
            if True, then nothing is reloaded but only data is loaded that is
            available as cached objects (not recommended to use but may be
            used if working offline without connection to database)
        filter_post : dict, optional
            filters applied to `UngriddedData` object AFTER it is read into
            memory, via :func:`UngriddedData.apply_filters`. This option was
            introduced in pyaerocom version 0.10.0 and should be used
            preferably over **kwargs. There is a certain flexibility with
            respect to how these filters can be defined, for instance, sub
            dicts for each `data_id`. The most common way would be
            to provide directly the input needed for
            `UngriddedData.apply_filters`. If you want to read multiple variables
            from one or more datasets, and if you want to apply variable
            specific filters, it is recommended to read the data individually
            for each variable and corresponding set of filters and then
            merge the individual filtered `UngriddedData` objects afterwards,
            e.g. using `data_var1 & data_var2`.
        **kwargs
            Additional input options for reading of data, which are applied
            WHILE the data is read. If any such additional options are
            provided that are applied during the reading, then automatic
            caching of the output `UngriddedData` object will be deactivated.
            Thus, it is recommended to handle data filtering via `filter_post`
            argument whenever possible, which will result in better performance
            as the unconstrained original data is read in and cached, and then
            the filtering is applied.

        Returns
        --------
        UngriddedData
            data object
        """
        force_caching = False
        if "force_caching" in kwargs:
            force_caching = kwargs.pop("force_caching")

        _caching = None
        if len(kwargs) > 0 and not force_caching:
            _caching = const.CACHING
            const.CACHING = False

            logger.info("Received additional reading constraints, ignoring caching")

        reader = self.get_lowlevel_reader(data_id)

        if vars_to_retrieve is None:
            vars_to_retrieve = reader.DEFAULT_VARS

        vars_to_retrieve = varlist_aerocom(vars_to_retrieve)

        # Since this interface enables to load multiple datasets, each of
        # which support a number of variables, here, only the variables are
        # considered that are supported by the dataset
        vars_available = [var for var in vars_to_retrieve if reader.var_supported(var)]

        if len(vars_available) == 0:
            raise DataRetrievalError(
                f"None of the input variables ({vars_to_retrieve}) is "
                f"supported by {data_id} interface"
            )
        cache = CacheHandlerUngridded(reader)
        if not self.ignore_cache:
            # initate cache handler
            for var in vars_available:
                try:
                    cache.check_and_load(var, force_use_outdated=only_cached)
                except Exception:
                    logger.exception(
                        "Fatal: compatibility error between old cache file "
                        "and current version of code."
                    )

        if not only_cached:
            vars_to_read = [v for v in vars_available if v not in cache.loaded_data]
        else:
            vars_to_read = []

        data_read = None
        if len(vars_to_read) > 0:
            _loglevel = logger.level
            logger.setLevel(logging.INFO)
            data_read = reader.read(vars_to_read, **kwargs)
            logger.setLevel(_loglevel)

            for var in vars_to_read:
                # write the cache file
                if not self.ignore_cache:
                    try:
                        cache.write(data_read, var)
                    except Exception as e:
                        _caching = False
                        logger.warning(
                            f"Failed to write to cache directory. "
                            f"Error: {repr(e)}. Deactivating caching "
                            f"in pyaerocom"
                        )

        if len(vars_to_read) == len(vars_available):
            data_out = data_read
        else:
            data_out = UngriddedData()
            for var in vars_available:
                if var in cache.loaded_data:
                    data_out.append(cache.loaded_data[var])
            if data_read is not None:
                data_out.append(data_read)

        # close the cache-object, keeps otherwise data-references
        cache = None

        if _caching is not None:
            const.CACHING = _caching

        if filter_post:
            filters = self._eval_filter_post(filter_post, data_id, vars_available)
            data_out = data_out.apply_filters(**filters)

        # Check to see if this reader is for a VerticalProfile
        # It is currently only allowed that a reader can be for a VerticalProfile, not a species
        if getattr(reader, "is_vertical_profile", None):
            data_out.is_vertical_profile = reader.is_vertical_profile

        return data_out

    def _eval_filter_post(self, filter_post, data_id, vars_available):
        filters = {}
        if not isinstance(filter_post, dict):
            raise ValueError(f"input filter_post must be dict, got {type(filter_post)}")
        elif len(filter_post) == 0:
            return filters

        if data_id in filter_post:
            # filters are specified specifically for that dataset
            subset = filter_post[data_id]
            return self._eval_filter_post(subset, data_id, vars_available)

        for key, val in filter_post.items():
            if key == "ignore_station_names":  # for backwards compatibility
                if isinstance(val, str | list):
                    filters["station_name"] = val
                    if "negate" not in filters:
                        filters["negate"] = []
                    filters["negate"].append("station_name")

                elif isinstance(val, dict):  # variable specific station filtering
                    if len(vars_available) > 1:
                        raise NotImplementedError(
                            f"Cannot filter different sites for multivariable "
                            f"UngriddedData objects (i.e. apply filter "
                            f"ignore_station_names={val} for UngriddedData "
                            f"object containing {vars_available}"
                        )
                    else:
                        # the variable that is available in the UngriddedData
                        # object
                        var = vars_available[0]
                        try:
                            filters["station_name"] = val[var]
                            if "negate" not in filters:
                                filters["negate"] = []
                            filters["negate"].append("station_name")
                        except KeyError:
                            continue
                else:
                    raise ValueError(f"Invalid input for ignore_station_names: {val}")
            else:
                filters[key] = val
        return filters

    def read_dataset_post(
        self,
        data_id,
        vars_to_retrieve,
        only_cached=False,
        filter_post=None,
        **kwargs,
    ):
        """Read dataset into an instance of :class:`ReadUngridded`

        Parameters
        ----------
        data_id : str
            name of dataset
        vars_to_retrieve : list
            variable or list of variables to be imported
        only_cached : bool
            if True, then nothing is reloaded but only data is loaded that is
            available as cached objects (not recommended to use but may be
            used if working offline without connection to database)
        filter_post : dict, optional
            filters applied to `UngriddedData` object AFTER it is read into
            memory, via :func:`UngriddedData.apply_filters`. This option was
            introduced in pyaerocom version 0.10.0 and should be used
            preferably over **kwargs. There is a certain flexibility with
            respect to how these filters can be defined, for instance, sub
            dicts for each `data_id`. The most common way would be
            to provide directly the input needed for
            `UngriddedData.apply_filters`. If you want to read multiple variables
            from one or more datasets, and if you want to apply variable
            specific filters, it is recommended to read the data individually
            for each variable and corresponding set of filters and then
            merge the individual filtered `UngriddedData` objects afterwards,
            e.g. using `data_var1 & data_var2`.
        **kwargs
            Additional input options for reading of data, which are applied
            WHILE the data is read. If any such additional options are
            provided that are applied during the reading, then automatic
            caching of the output `UngriddedData` object will be deactivated.
            Thus, it is recommended to handle data filtering via `filter_post`
            argument whenever possible, which will result in better performance
            as the unconstrained original data is read in and cached, and then
            the filtering is applied.

        Returns
        --------
        UngriddedData
            data object
        """
        aux_info = self.post_compute[data_id]
        loaded = []
        for var in vars_to_retrieve:
            input_data_ids_vars = []
            aux_info_var = aux_info["aux_requires"][var]
            for aux_id, aux_vars in aux_info_var.items():
                if aux_id in self.post_compute:
                    aux_data = self.read_dataset_post(
                        data_id=aux_id,
                        vars_to_retrieve=aux_vars,
                        only_cached=only_cached,
                        **kwargs,
                    )
                    for aux_var in aux_vars:
                        input_data_ids_vars.append((aux_data, aux_id, aux_var))
                else:
                    # read variables individually, so filter_post is more
                    # flexible if some post filters are specified for
                    # individual variables...
                    for aux_var in aux_vars:
                        _data = self.read_dataset(
                            data_id=aux_id,
                            vars_to_retrieve=aux_var,
                            only_cached=only_cached,
                            filter_post=filter_post,
                            **kwargs,
                        )
                        input_data_ids_vars.append((_data, aux_id, aux_var))

            aux_merge_how = aux_info["aux_merge_how"][var]

            if var in aux_info["aux_units"]:
                var_unit_out = aux_info["aux_units"][var]
            else:
                var_unit_out = None

            if aux_merge_how == "eval":
                # function MUST be defined
                aux_fun = aux_info["aux_funs"][var]
            else:
                aux_fun = None

            merged_stats = combine_vardata_ungridded(
                data_ids_and_vars=input_data_ids_vars,
                merge_eval_fun=aux_fun,
                merge_how=aux_merge_how,
                var_name_out=var,
                var_unit_out=var_unit_out,
                data_id_out=aux_info["data_id"],
            )
            loaded.append(UngriddedData.from_station_data(merged_stats))
        first = loaded[0]
        if len(loaded) == 1:
            return first
        for data in loaded[1:]:
            first.append(data)
        return first

    def read(
        self,
        data_ids=None,
        vars_to_retrieve=None,
        only_cached=False,
        filter_post=None,
        configs: PyaroConfig | list[PyaroConfig] | None = None,
        **kwargs,
    ):
        """Read observations

        Iter over all datasets in :attr:`data_ids`, call
        :func:`read_dataset` and append to data object

        Parameters
        ----------
        data_ids : str or list
            data ID or list of all datasets to be imported
        vars_to_retrieve : str or list
            variable or list of variables to be imported
        only_cached : bool
            if True, then nothing is reloaded but only data is loaded that is
            available as cached objects (not recommended to use but may be
            used if working offline without connection to database)
        filter_post : dict, optional
            filters applied to `UngriddedData` object AFTER it is read into
            memory, via :func:`UngriddedData.apply_filters`. This option was
            introduced in pyaerocom version 0.10.0 and should be used
            preferably over **kwargs. There is a certain flexibility with
            respect to how these filters can be defined, for instance, sub
            dicts for each `data_id`. The most common way would be
            to provide directly the input needed for
            `UngriddedData.apply_filters`. If you want to read multiple variables
            from one or more datasets, and if you want to apply variable
            specific filters, it is recommended to read the data individually
            for each variable and corresponding set of filters and then
            merge the individual filtered `UngriddedData` objects afterwards,
            e.g. using `data_var1 & data_var2`.
        **kwargs
            Additional input options for reading of data, which are applied
            WHILE the data is read. If any such additional options are
            provided that are applied during the reading, then automatic
            caching of the output `UngriddedData` object will be deactivated.
            Thus, it is recommended to handle data filtering via `filter_post`
            argument whenever possible, which will result in better performance
            as the unconstrained original data is read in and cached, and then
            the filtering is applied.

        Example
        -------
        >>> import pyaerocom.io.readungridded as pio
        >>> from pyaerocom import const
        >>> obj = pio.ReadUngridded(data_id=const.AERONET_SUN_V3L15_AOD_ALL_POINTS_NAME)
        >>> obj.read()
        >>> print(obj)
        >>> print(obj.metadata[0.]['latitude'])

        """
        if data_ids is None:
            data_ids = self.data_ids
        elif isinstance(data_ids, str):
            data_ids = [data_ids]

        if configs is not None:
            if not isinstance(configs, list):
                configs = [configs]
            for config in configs:
                self._init_pyaro_reader(config=config)
                data_ids.append(config.name)

        if isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        data = UngriddedData()
        for ds in data_ids:
            if ds in self.post_compute:
                data.append(
                    self.read_dataset_post(
                        data_id=ds,
                        vars_to_retrieve=vars_to_retrieve,
                        only_cached=only_cached,
                        filter_post=filter_post,
                        **kwargs,
                    )
                )
            else:
                data_to_append = self.read_dataset(
                    data_id=ds,
                    vars_to_retrieve=vars_to_retrieve,
                    only_cached=only_cached,
                    filter_post=filter_post,
                    **kwargs,
                )
                data.append(data_to_append)
                # TODO: Test this. UngriddedData can contain more than 1 variable
                if getattr(data_to_append, "is_vertical_profile", None):
                    data.is_vertical_profile = data_to_append.is_vertical_profile

            logger.info(f"Successfully imported {ds} data")
        return data

    def _check_var_alias(self, var, supported):
        # could be an alias
        aliases = get_aliases(var)
        for svar in supported:
            if svar in aliases:
                return svar
        raise ValueError()

    def get_vars_supported(self, obs_id, vars_desired):  # , config: Optional[PyaroConfig] = None):
        """
        Filter input list of variables by supported ones for a certain data ID

        Parameters
        ----------
        obs_id : str
            ID of observation network
        vars_desired : list
            List of variables that are desired

        Returns
        -------
        list
            list of variables that can be read through the input network

        """
        obs_vars = []
        if isinstance(vars_desired, str):
            vars_desired = [vars_desired]
        if obs_id in self.post_compute:
            # check if all required are accessible
            postinfo = self.post_compute[obs_id]
            supported = postinfo["vars_supported"]
            for var in varlist_aerocom(vars_desired):
                if var not in supported:
                    try:
                        var = self._check_var_alias(var, supported)
                    except ValueError:
                        # no alias match, skip...
                        continue
                requires = postinfo["aux_requires"][var]
                all_good = True
                for ds, vars_required in requires.items():
                    if isinstance(vars_required, str):
                        vars_required = [vars_required]
                    vars_avail = self.get_vars_supported(ds, vars_required)
                    if not len(vars_required) == len(vars_avail):
                        all_good = False
                        break
                if all_good:
                    obs_vars.append(var)

        else:
            # check if variable can be read from a dataset on disk
            _oreader = self.get_lowlevel_reader(obs_id)
            for var in varlist_aerocom(vars_desired):
                if _oreader.var_supported(var):
                    obs_vars.append(var)
        return obs_vars

    def __str__(self):
        return "\n".join(str(self.get_lowlevel_reader(ds)) for ds in self.data_ids)



import pytest

from pyaerocom import const
from pyaerocom.data import resources
from pyaerocom.exceptions import VariableDefinitionError
from pyaerocom.varcollection import VarCollection
from pyaerocom.variable import Variable


def test_VARS_is_VarCollection():
    assert isinstance(const.VARS, VarCollection)


@pytest.fixture()
def collection() -> VarCollection:
    with resources.path("pyaerocom.data", "variables.ini") as path:
        assert path.exists()
        return VarCollection(str(path))


def test_invalid_entries(collection: VarCollection):
    invalid = ("_" in var for var in collection.all_vars)
    assert sum(invalid) == 0


@pytest.mark.parametrize(
    "var_ini,exception,error",
    [
        pytest.param(None, ValueError, "Invalid input for var_ini, need str", id="ValueError"),
        pytest.param(
            "/bla/blub",
            FileNotFoundError,
            "File /bla/blub does not exist",
            id="FileNotFoundError",
        ),
    ],
)
def test_VarCollection___init___error(var_ini, exception, error: str):
    with pytest.raises(exception) as e:
        VarCollection(var_ini)
    assert str(e.value) == error


def test_VarCollection_add_var(collection: VarCollection):
    var = Variable(var_name="concpm10gt1", units="ug m-3")
    collection.add_var(var)
    assert var.var_name in collection.all_vars


def test_VarCollection_add_var_error(collection: VarCollection):
    var = Variable(var_name="concpm10", units="ug m-3")
    with pytest.raises(VariableDefinitionError) as e:
        collection.add_var(var)
    assert str(e.value) == f"variable with name {var.var_name} is already defined"


def test_VarCollection_delete_var(collection: VarCollection):
    var = Variable(var_name="concpm10", units="ug m-3")
    collection.delete_variable(var.var_name)
    assert var.var_name not in collection.all_vars


def test_VarCollection_delete_var_error(collection: VarCollection):
    var = Variable(var_name="concpm10gt1", units="ug m-3")
    with pytest.raises(VariableDefinitionError) as e:
        collection.delete_variable(var.var_name)
    assert str(e.value) == f"No such variable {var.var_name} in VarCollection"


@pytest.mark.parametrize("var_name", ["blablub42", "od550aer"])
def test_VarCollection_get_var(collection: VarCollection, var_name: str):
    collection.add_var(Variable(var_name="blablub42"))
    assert isinstance(collection.get_var(var_name), Variable)


def test_VarCollection_get_var_error(collection: VarCollection):
    var_name = "bla"
    with pytest.raises(VariableDefinitionError) as e:
        collection.get_var(var_name)
    assert str(e.value) == f"Error (VarCollection): input variable {var_name} is not supported"


@pytest.mark.parametrize(
    "search_pattern,num",
    [
        ("*blaaaaaaa*", 0),
        ("dep*", 9),
        ("od*", 26),
        ("conc*", 104),
    ],
)
def test_VarCollection_find(collection: VarCollection, search_pattern: str, num: int):
    result = collection.find(search_pattern)
    assert len(result) == num


def test_VarCollection_delete_var_MULTIDEF(collection: VarCollection):
    var_name = "concpm10"
    collection.all_vars.append(var_name)
    with pytest.raises(VariableDefinitionError) as e:
        collection.delete_variable(var_name)
    assert f"found multiple matches for variable {var_name} in VarCollection" in str(e.value)


def test_VarCollection___dir__(collection: VarCollection):
    result = dir(collection)
    all_vars = collection.all_vars

    assert len(result) == len(all_vars)
    assert result == sorted(all_vars)


@pytest.mark.parametrize("var_name,found", [("blablub", False), ("od550aer", True)])
def test_VarCollection___contains__(collection: VarCollection, var_name: str, found: bool):
    assert (var_name in collection) == found


def test_VarCollection___len__(collection: VarCollection):
    assert len(collection) > 0


def test_VarCollection___getitem__(collection: VarCollection):
    assert isinstance(collection["od550aer"], Variable)


def test_VarCollection___getitem___error(collection: VarCollection):
    var_name = "blaaaa"
    assert var_name not in collection
    with pytest.raises(VariableDefinitionError) as e:
        collection[var_name]
    assert str(e.value).endswith(f"input variable {var_name} is not supported")


def test_VarCollection___repr__(collection: VarCollection):
    assert repr(collection).startswith("VarCollection")


def test_VarCollection___str__(collection: VarCollection):
    assert str(collection).startswith("VarCollection")



from __future__ import annotations

from datetime import date, datetime, timedelta
from pathlib import Path

import pytest

from pyaerocom import const
from pyaerocom.io.cams2_83.read_obs import DATA_FOLDER_PATH, ReadCAMS2_83
from pyaerocom.io.cams2_83.read_obs import obs_paths as find_obs_paths
from pyaerocom.io.readungridded import ReadUngridded
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.ungriddeddata import UngriddedData

TEST_DATE = datetime(2021, 12, 1)
TEST_DATES = [TEST_DATE + timedelta(days=d) for d in range(3)]


@pytest.fixture
def obs_paths(dates: list[str | date | datetime]) -> list[Path]:
    if not DATA_FOLDER_PATH.is_dir():
        pytest.skip(f"no access to {DATA_FOLDER_PATH}")
    paths = find_obs_paths(*dates)
    return list(paths)


@pytest.mark.parametrize("dates", [TEST_DATES])
def test_obs_paths(obs_paths: list[Path]):
    for path in obs_paths:
        assert path.exists()


def test_init():
    data = ReadCAMS2_83()
    assert isinstance(data, ReadUngriddedBase)


@pytest.mark.parametrize("dates", [TEST_DATES])
def test_read_ungridded(obs_paths: list[Path]):
    data = ReadUngridded().read(const.CAMS2_83_NRT_NAME, "concco", files=obs_paths)
    assert isinstance(data, UngriddedData)
